structure Alloc :> ALLOC =
struct

  exception Convert of string
  exception Error of string

  structure Var = Variable

  fun remove_useless_bnd x v bs body repl =
    case v of
      IL7.ValLit l => (bs@[IL7.BndLit(x, l)], body)
    | IL7.ValVar y =>
      let
        val body = repl x y body
        val bs = map (IL7.replace_exp_bnd x y) bs
      in
        (bs, body)
      end

  fun convert_word x w body repl=
    let
      val x' = Var.freshen x
    in
      case w of
        IL6.WProd vs =>
         let
           val (bsl, vs') = ListPair.unzip (map convert_value vs)
           val (_, ivs) =
              foldl (fn (v, (i, l)) => (i + 1, (i, v)::l)) (1, []) vs'
           fun make_assn (i, v) = IL7.BndSet(x', i, v)
           val bnds = map make_assn (rev ivs)
           val body = repl x x' body
         in
           ((List.concat bsl) @ (IL7.BndAlloc(x', length vs')::bnds), body)
         end
      | IL6.WRoll(c, v) =>
         let val (bs, v') = convert_value v
             val body = repl x x' body
             val (bout, body) = remove_useless_bnd x' v' [] body repl
         in (bs@bout, body)
         end
      | IL6.WSumInj(c, i, v) =>
         let
           val (bs, v') = convert_value v
           val body = repl x x' body
           val bndsum =
             [IL7.BndAlloc(x', 2),
              IL7.BndSet(x', 1, IL7.valInt i),
              IL7.BndSet(x', 2, v')
             ]
         in
           (bs@bndsum, body)
         end
      | IL6.WPack((c1, v), c2) =>
         let val (bs, v') = convert_value v
             val body = repl x x' body
             val (bout, body) = remove_useless_bnd x' v' [] body repl
         in (bs@bout, body)
         end
      | IL6.WTagInj(v1, v2) =>
         let
           val (bs1, v1') = convert_value v1
           val (bs2, v2') = convert_value v2
           val bndsum =
             [IL7.BndAlloc(x', 2),
              IL7.BndSet(x', 1, v1'),
              IL7.BndSet(x', 2, v2')
             ]
           val body = repl x x' body
         in
           (bs1@bs2@bndsum, body)
         end
      | IL6.WRef v =>
        let
          val body = repl x x' body
          val (bs, v') = convert_value v
          val bndref =
            [IL7.BndAlloc(x', 1),
             IL7.BndSet(x', 1, v')]
        in
          (bs@bndref, body)
        end
      | IL6.WNewTag c => ([IL7.BndNewTag x'], repl x x' body)
    end

  and convert_bnd b body repl=
    case b of
      IL6.BndVal(x, v) =>
        let
          val x' = Var.freshen x
          val body = repl x x' body
          val (bs, v') = convert_value v
          val (bout, body) = remove_useless_bnd x' v' [] body repl
        in
          (bs@bout, body)
        end
    | IL6.BndAlloc (x, w) => convert_word x w body repl
    | IL6.BndUnOp(x, oper, v) =>
        let
          val x' = Var.freshen x
          val (bs, v') = convert_value v
          val body = repl x x' body
        in
          (bs@[IL7.BndUnOp(x', oper, v')], body)
        end
    | IL6.BndBinOp(x, v1, oper, v2) =>
        let
          val x' = Var.freshen x
          val (bs1, v1') = convert_value v1
          val (bs2, v2') = convert_value v2
          val body = repl x x' body
        in
          (bs1@bs2@[IL7.BndBinOp(x', v1', oper, v2')], body)
        end
    | IL6.BndGet(x, v) =>
        let
          val x' = Var.freshen x
          val (bs, v') = convert_value v
          val body = repl x x' body
        in
          (bs@[IL7.BndProj(x', 1, v')], body)
        end
    | IL6.BndSet(x, v1, v2) =>
        let
          val x' = Var.freshen x
          val t = Var.newNamed "t"
          val tvar = IL7.ValVar t
          val (bs1, v1') = convert_value v1
          val (bs2, v2') = convert_value v2
          val body = repl x x' body
          val bnds = [IL7.BndSet(t, 1, v2'), IL7.bndUnit x']
          val (bset, body) =
                remove_useless_bnd t v1' bnds  body repl
        in
          (bs1@bs2@bset, body)
        end
    | IL6.BndProj(x, i, v) =>
        let
          val x' = Var.freshen x
          val (bs, v') = convert_value v
          val body = repl x x' body
        in
          (bs@[IL7.BndProj(x', i, v')], body)
        end
    | IL6.BndUnroll(x, v) =>
        let
          val x' = Var.freshen x
          val (bs, v') = convert_value v
          val body = repl x x' body
          val (bset, body) = remove_useless_bnd x' v' [] body repl
        in
          (bs@bset, body)
        end
    
  and convert_value v =
    case v of
      IL6.ValVar x => ([], IL7.ValVar x)
    | IL6.ValLit l => ([], IL7.ValLit l)
    | IL6.ValConAbs(a, k, v) => convert_value v
    | IL6.ValConApp(v, c) => convert_value v
    | IL6.ValLet(b, v) =>
        let
          val (bsv, v') = convert_value v
          (* We use halt to turn this into an expression
            since we couldn't get convert_bnd to be properly
            polymorphic over values and expressions *)
          val (bs, v') =
            case convert_bnd b (IL7.ExHalt v') IL7.replace_exp_exp of
              (bs, IL7.ExHalt v') => (bs, v')
            | _ => raise Convert "Didn't get help out of convert"
        in
          (bs@bsv, v')
        end

  fun convert_exp e =
    case e of
      IL6.ExHalt v =>
        let val (bs, v') = convert_value v
        in
          foldr IL7.ExLet (IL7.ExHalt v') bs
        end
    | IL6.ExApp(v1, v2) =>
        let val (bs1, v1') = convert_value v1
            val (bs2, v2') = convert_value v2
        in
          foldr IL7.ExLet (IL7.ExApp(v1', v2')) (bs1 @ bs2)
        end
    | IL6.ExCase(v, arms) =>
        let
          val tag = Var.newNamed "tag"
          val tagvar = IL7.ValVar tag
          val x = Var.newNamed "x"
          val (bsv, v') = convert_value v
          fun convert_case_arm (xi, ei) =
              convert_exp(IL6.replace_exp_exp xi x ei)
          val armsout = map convert_case_arm arms
        in
            IL7.ExLet(IL7.BndProj(tag, 1, v'),
            IL7.ExLet(IL7.BndProj(x, 2, v'),
              IL7.ExSwitch(tagvar, armsout)))
        end
    | IL6.ExIfTagOf(v1, v2, (x, e1), e2) =>
        let
          val tag = Var.newNamed "tag"
          val res = Var.newNamed "res"
          val teq = Var.newNamed "teq"
          val tagvar = IL7.ValVar tag
          val resvar = IL7.ValVar res
          val teqvar = IL7.ValVar teq
          val x' = Var.freshen x
          val (bsv1, v1') = convert_value v1
          val (bsv2, v2') = convert_value v2
          val e1' = IL6.replace_exp_exp x x' e1
          val e1' = convert_exp e1'
          val e2' = convert_exp e2
          val sbody =
            IL7.ExLet(IL7.BndProj(tag, 1, v1'),
            IL7.ExLet(IL7.BndBinOp(teq, tagvar, Literal.OpEquals, v2'),
            IL7.ExLet(IL7.BndProj(res, 1, teqvar),
              IL7.ExSwitch(resvar,
                [IL7.ExLet(IL7.BndProj(x', 2, v1'), e1'), e2']))))
        in
          foldr IL7.ExLet sbody (bsv1 @ bsv2)
        end
    | IL6.ExLet(b, e) =>
      let
        val e' = convert_exp e
        val (bs, e') = convert_bnd b e' IL7.replace_exp_exp
      in
        foldr IL7.ExLet e' bs
      end
    | IL6.ExUnpack(a, x, v, e) =>
        let
            val (a', e') = IL6.alpha_vary (a, e) IL6.replace_con_exp 
            val (x', e') = IL6.alpha_vary (x, e') IL6.replace_exp_exp
            val (bs, v') = convert_value v
            val e' = convert_exp e'
            val (bset, e') = remove_useless_bnd x' v' [] e' IL7.replace_exp_exp
        in
          foldr IL7.ExLet e' (bs@bset)
        end

  fun convert_program (fbnds, e) =
    let
        fun convert_fbnd (f, (cvardecs, x, c, e)) =
          let
            val (x', e') = IL6.alpha_vary (x, e) IL6.replace_exp_exp
            val eout = convert_exp e'
          in
            (f, (x', eout))
          end
        val fbnds' = map convert_fbnd fbnds
        val e' = convert_exp e
    in
      (fbnds', e')
    end
end
