(* Adds a convert_program function to the CPS_CONVERT signature
 * This allows us to write a compiler specific top level
 * continutation and exception handler. We do this separately
 * from the CPS_CONVERT signature because it is some sense fairly
 * arbitrary and thus outside the scope of the cps conversion
 * assignment
 *)

signature CPS_ADAPTOR =
sig
  include CPS_CONVERT
  val convert_program : context -> IL3.exp -> IL4.exp
end

functor CPSAdaptorFn(Converter : CPS_CONVERT) :> CPS_ADAPTOR =
struct
    open Converter
    fun convert_program ctx exp =
        let
          val f = Variable.newNamed "f"
          val x = Variable.newNamed "x"
          val xvar = IL4.ValVar x
          val (c, {k, k_exn}, e) = convert_exp ctx exp
        in
          IL4.ExLet(IL4.BndVal(k,
            IL4.ValFix(1, [(f, x, convert_con c, IL4.ExHalt xvar)])),
          IL4.ExLet(IL4.BndVal(k_exn,
            IL4.ValFix(1,
              [(f, x, IL4.TyTagged,
                IL4.ExLet(
                  IL4.BndUnOp(
                    x,
                    Literal.OpPrint,
                    IL4.valString "uncaught exception...exiting\n"),
                  IL4.ExHalt IL4.valUnit))])), e))
        end
end
