structure Hoist :> HOIST =
  struct
    exception Convert of string
    exception Error of string

    structure IL5S = IL5Static
    structure Ctx = IL5S.Ctx
    structure Var = Variable

    type context = Ctx.context

    (* this is used to postpone the concatenation of
     all the fbnds until the end of the process *)
    datatype 'a ltree = B of 'a ltree list | L of 'a

    fun flatten (L x) acc = x::acc
      | flatten (B ls) acc = (foldr (fn (l, f) => fn acc => f (flatten l acc))
                                   (fn x => x) ls) acc

    fun convert_kind k =
      case k of
        IL5.KndType => IL6.KndType
      | IL5.KndArrow(k1, k2) => IL6.KndArrow(convert_kind k1, convert_kind k2)
      | IL5.KndUnit => IL6.KndUnit
      | IL5.KndPair(k1, k2) => IL6.KndPair(convert_kind k1, convert_kind k2)

    fun convert_con c =
      case c of
        IL5.ConVar a => IL6.ConVar a
      | IL5.ConAbs(a, k, c) => IL6.ConAbs(a, convert_kind k, convert_con c)
      | IL5.ConApp(c1, c2) => IL6.ConApp(convert_con c1, convert_con c2)
      | IL5.ConUnit => IL6.ConUnit
      | IL5.ConPair(c1, c2) => IL6.ConPair(convert_con c1, convert_con c2)
      | IL5.ConProj1 c => IL6.ConProj1 (convert_con c)
      | IL5.ConProj2 c => IL6.ConProj2 (convert_con c)
      | IL5.ConMu(a, k, c) => IL6.ConMu(a, convert_kind k, convert_con c)
      | IL5.TyLit l => IL6.TyLit l
      | IL5.TyCont c => IL6.TyCont([], convert_con c)
      | IL5.TyRef c => IL6.TyRef(convert_con c)
      | IL5.TyTagged => IL6.TyTagged
      | IL5.TyTag c => IL6.TyTag(convert_con c)
      | IL5.TySum arms => IL6.TySum(map convert_con arms)
      | IL5.TyProd arms => IL6.TyProd(map convert_con arms)
      | IL5.TyForall(a, k, c) =>
          IL6.TyForall(a, convert_kind k, convert_con c)
      | IL5.TyExists(a, k, c) =>
          IL6.TyExists(a, convert_kind k, convert_con c)

    fun convert_bnd G b =
      case b of
        IL5.BndVal(x, v) =>
          let val (fs, v') = convert_value G v
              val x' = Var.freshen x
          in
            (fs, x, x', IL5S.synth_value_con G v, IL6.BndVal(x', v'))
          end
      | IL5.BndUnOp(x, oper, v) =>
          let val (fs, v') = convert_value G v
              val x' = Var.freshen x
              val (_, cout) =
                    Literal.unop_tys oper IL5.TyLit IL5.tyBool IL5.tyUnit
          in
            (fs, x, x', cout, IL6.BndUnOp(x', oper, v'))
          end
      | IL5.BndBinOp(x, v1, oper, v2) =>
          let val (fs1, v1') = convert_value G v1
              val (fs2, v2') = convert_value G v2
              val x' = Var.freshen x
              val (_, _, cout) =
                    Literal.binop_tys oper IL5.TyLit IL5.tyBool IL5.tyUnit
          in
            (B[fs1, fs2], x, x', cout, IL6.BndBinOp(x', v1', oper, v2'))
          end
      | IL5.BndRef(x, v) =>
          let val (fs, v') = convert_value G v
              val x' = Var.freshen x
          in
            (fs, x, x', IL5.TyRef (IL5S.synth_value_con G v),
              IL6.BndAlloc(x', IL6.WRef v'))
          end
      | IL5.BndGet(x, v) =>
          let val (fs, v') = convert_value G v
              val x' = Var.freshen x
              val c = case IL5S.whnf (IL5S.synth_value_con G v) of
                           IL5.TyRef c => c
                         | _ => raise Convert "non ref value in deref"
          in
            (fs, x, x',c, IL6.BndGet(x', v'))
          end
      | IL5.BndSet(x, v1, v2) =>
          let val (fs1, v1') = convert_value G v1
              val (fs2, v2') = convert_value G v2
              val x' = Var.freshen x
          in (B[fs1, fs2], x, x', IL5.tyUnit, IL6.BndSet(x', v1', v2'))
          end
      | IL5.BndProj(x, i, v) =>
          let val (fs, v') = convert_value G v
              val x' = Var.freshen x
              val cs = case IL5S.synth_value_con G v of
                           IL5.TyProd cs => cs
                         | _ => raise Convert "non product in projection"
          in
             (fs, x, x', List.nth(cs, i - 1), IL6.BndProj(x', i, v'))
          end
      | IL5.BndNewTag(x, c) =>
          let val x' = Var.freshen x
          in
            (B [], x, x', IL5.TyTag c,
              IL6.BndAlloc(x', IL6.WNewTag (convert_con c)))
          end
      | IL5.BndUnroll(x, v) =>
          let val (fs, v') = convert_value G v
              val x' = Var.freshen x
              val c = (IL5S.whnf (IL5S.synth_value_con G v))
          in (fs, x, x', IL5S.expand G c, IL6.BndUnroll(x', v'))
          end

    and convert_value G value =
      case value of
        IL5.ValVar x => (B [], IL6.ValVar x)
      | IL5.ValLit l => (B [], IL6.ValLit l)
      | IL5.ValFix(i, fbnds) =>
          let
            val fbnds' = map (fn (f, x, c, e) =>
                (f, Var.freshen f, x, Var.freshen x, c, e)) fbnds
            fun vary_fbnd fbnds (f, g, x, y, c, e) =
              let
                val re = IL5.replace_exp_exp
                val e' = foldl (fn ((f, g, _, _, _, _), e) => re f g e) e fbnds
                val e'' = re x y e'
              in
                (g, y, c, e'')
              end
            val fbnds' = map (vary_fbnd fbnds') fbnds'
            val D = Ctx.convar_component G
            val ds =
              map (fn Ctx.DecConKind(a, k) => (a, k)
                    | _ => raise Error
                            "got non-convar out of convar_component")
                  (Ctx.to_list D)
            val dsout = rev (map (fn (a, k) => (a, convert_kind k)) ds)
            val fdecsl =
              map (fn (g, _, c, _) => Ctx.DecExpCon(g, IL5.TyCont c)) fbnds'
            val fdecs = foldl (op::) [] fdecsl
            val G' = Ctx.extend_many G fdecs
            fun convert_fbnd(g, y, c, e) =
              let
                (* Applies the tyvars to recursive uses of the functions *)
                fun fix_body ((g, y, c, ei), e) =
                  let
                    val g' = Var.freshen g
                    val e = IL6.replace_exp_exp g g' e
                    val gbody = foldr (fn((c, _), e) =>
                          IL6.ValConApp(e, IL6.ConVar c)) (IL6.ValVar g) ds
                  in
                    IL6.ExLet(IL6.BndVal(g', gbody), e)
                  end
                val G'' = Ctx.extend G' (Ctx.DecExpCon(y, c))
                val (fs, e') = convert_exp' G'' e
                val e' = foldl fix_body e' fbnds'
                val c' = convert_con c
              in
                (fs, (g, (dsout, y, c', e')))
              end
            val (fsoutl, fbnds') = ListPair.unzip(map convert_fbnd fbnds')
            val fsout = B fsoutl
            val (f, _) = List.nth(fbnds', i - 1)
            val outval =
                  foldr (fn((c, _), e) =>
                          IL6.ValConApp(e, IL6.ConVar c)) (IL6.ValVar f) ds
          in
            (B[fsout, B (map L fbnds')], outval)
          end
      | IL5.ValProd arms =>
          let val (fs, elems) = ListPair.unzip (map (convert_value G) arms)
              val p = Var.newNamed "p"
              val outval =
                IL6.ValLet(IL6.BndAlloc(p, IL6.WProd elems), IL6.ValVar p)
          in
            (B fs, outval)
          end
      | IL5.ValSumInj(c, i, v) =>
          let
            val (fs, v') = convert_value G v
            val c' = convert_con c
            val s = Var.newNamed "s"
            val vout =
              IL6.ValLet(IL6.BndAlloc(s, IL6.WSumInj(c', i, v')), IL6.ValVar s)
          in
            (fs, vout)
          end
      | IL5.ValTagInj(v1, v2) =>
          let
            val (fs1, v1') = convert_value G v1
            val (fs2, v2') = convert_value G v2
            val t = Var.newNamed "t"
            val vout =
              IL6.ValLet(IL6.BndAlloc(t, IL6.WTagInj(v1', v2')), IL6.ValVar t)
          in
            (B[fs1, fs2], vout)
          end
      | IL5.ValConAbs(a, k, v) =>
          let
            val (a', v') = IL5.alpha_vary (a, v) IL5.replace_con_value
            val k' = convert_kind k
            val G' = Ctx.extend G (Ctx.DecConKind(a', k))
            val (fs, vout) = convert_value G' v'
          in
            (fs, IL6.ValConAbs(a', k', vout))
          end
      | IL5.ValConApp(v, c) =>
          let
            val (fs, v') = convert_value G v
            val c' = convert_con c
          in
            (fs, IL6.ValConApp(v', c'))
          end
      | IL5.ValRoll(c, v) =>
          let
            val (fs, v') = convert_value G v
            val c' = convert_con c
            val r = Var.newNamed "r"
            val vout =
              IL6.ValLet(IL6.BndAlloc(r, IL6.WRoll(c', v')), IL6.ValVar r)
          in
            (fs, vout)
          end
      | IL5.ValPack((c1, v), c2) =>
          let
            val (fs, v') = convert_value G v
            val c1' = convert_con c1
            val c2' = convert_con c2
            val p = Var.newNamed "p"
            val vout = IL6.ValLet(IL6.BndAlloc(p, IL6.WPack((c1', v'), c2')),
                          IL6.ValVar p)
          in
            (fs, vout)
          end
 
    and convert_exp' G exp =
      case exp of
        IL5.ExHalt v =>
          let
            val (fs, v') = convert_value G v
          in
            (fs, IL6.ExHalt v')
          end
      | IL5.ExApp(v1, v2) =>
          let
            val (fs1, v1') = convert_value G v1
            val (fs2, v2') = convert_value G v2
          in
            (B[fs1, fs2], IL6.ExApp(v1', v2'))
          end
      | IL5.ExCase(v, arms) =>
          let
            val (fsv, v') = convert_value G v
            val carms =
              case IL5S.whnf (IL5S.synth_value_con G v) of
                IL5.TySum carms => carms
              | _ => raise Error "casing on non-sum type"
            fun convert_arm (c, (x, e)) =
              let
                val (x', e') = IL5.alpha_vary (x, e) IL5.replace_exp_exp
                val G' = Ctx.extend G (Ctx.DecExpCon(x', c))
                val (fs, e') = convert_exp' G' e'
              in
                (fs, (x', e'))
              end
            val (farms, arms') =
              ListPair.unzip (map convert_arm (ListPair.zip(carms, arms)))
          in
            (B[fsv, B farms], IL6.ExCase(v', arms'))
          end
      | IL5.ExIfTagOf(v1, v2, (x, e1), e2) =>
          let
            val (fsv1, v1') = convert_value G v1
            val (fsv2, v2') = convert_value G v2
            val c =
              case IL5S.whnf (IL5S.synth_value_con G v2) of
                IL5.TyTag c => c
              | _ => raise Error "tag value with non tag type"
            val (x', e1') = IL5.alpha_vary (x, e1) IL5.replace_exp_exp
            val G' = Ctx.extend G (Ctx.DecExpCon(x', c))
            val (fse1, e1') = convert_exp' G' e1'
            val (fse2, e2') = convert_exp' G e2
          in
            (B[fsv1, fsv2, fse1, fse2], IL6.ExIfTagOf(v1', v2', (x', e1'), e2'))
          end
      | IL5.ExLet(b, e) =>
          let
            val (fsb, x, x', c, b') = convert_bnd G b
            val e' = IL5.replace_exp_exp x x' e
            val G' = Ctx.extend G (Ctx.DecExpCon(x', c))
            val (fse, e') = convert_exp' G' e'
          in
            (B[fsb, fse], IL6.ExLet(b', e'))
          end
      | IL5.ExUnpack(a, x, v, e) =>
          let
            val (fsv, v') = convert_value G v
            val (a', e') = IL5.alpha_vary (a, e) IL5.replace_con_exp
            val (x', e') = IL5.alpha_vary (x, e') IL5.replace_exp_exp
            val (b, k, c) =
              case IL5S.whnf (IL5S.synth_value_con G v) of
                IL5.TyExists x => x
              | _ => raise Error "unpacking value with non-existential type"
            val c' = IL5.replace_con_con b a' c
            val G' = Ctx.extend G (Ctx.DecExpCon(x', c'))
            val G' = Ctx.extend G' (Ctx.DecConKind(a', k))
            val (fse, e') = convert_exp' G' e'
          in
            (B[fse, fsv], IL6.ExUnpack(a', x', v', e'))
          end
   fun convert_exp G e =
    let
      val (fs, e') = convert_exp' G e
    in
      (flatten fs [], e')
    end
  end
