structure SingElim :> SING_ELIM =
  struct
    structure IL2S = IL2Static
    structure Ctx = IL2S.Ctx
    type context = Ctx.context

    structure Var = Variable

    exception Convert of string
    exception Error of string


    fun convert_kind k =
      case k of
        IL2.KndType => IL3.KndType
      | IL2.KndPi(a, k1, k2) => IL3.KndArrow(convert_kind k1, convert_kind k2)
      | IL2.KndUnit => IL3.KndUnit
      | IL2.KndSigma(a, k1, k2) => IL3.KndPair(convert_kind k1, convert_kind k2)
      | IL2.KndSing _ => IL3.KndType

    fun translate_con c k = 
          case k of
            IL2.KndType => c
          | IL2.KndSing c' => c'
          | IL2.KndUnit => IL2.ConUnit
          | IL2.KndPi(a, k1, k2) =>
            let
              val (a, k2) = IL2.alpha_vary (a, k2) IL2.replace_con_kind
              val at = translate_con (IL2.ConVar a) k1
              val ct = translate_con at k1
              val kt = IL2.subst_con_kind ct a k2
            in
              IL2.ConAbs(a, k1, translate_con (IL2.ConApp(c, ct)) kt)
            end
          | IL2.KndSigma(a, k1, k2) =>
             let
               val c1out = translate_con (IL2.ConProj1 c) k1
               val k2t = IL2.subst_con_kind c1out a k2
               val c2out = translate_con (IL2.ConProj2 c) k2t
             in
               IL2.ConPair(c1out, c2out)
             end

    fun convert_ctx G =
      List.mapPartial
        (fn (Ctx.DecConKind(a, k)) => SOME(a, translate_con (IL2.ConVar a) k)
            | _ => NONE) (Ctx.to_list G)

    fun erase_con c =
      case c of
        IL2.ConVar a => IL3.ConVar a
      | IL2.ConAbs(a, k, c) => IL3.ConAbs(a, convert_kind k, erase_con c)
      | IL2.ConApp(c1, c2) => IL3.ConApp(erase_con c1, erase_con c2)
      | IL2.ConUnit => IL3.ConUnit
      | IL2.ConPair(c1, c2) => IL3.ConPair(erase_con c1, erase_con c2)
      | IL2.ConProj1 c => IL3.ConProj1(erase_con c)
      | IL2.ConProj2 c => IL3.ConProj2(erase_con c)
      | IL2.ConMu(a, k, c) => IL3.ConMu(a, convert_kind k, erase_con c)
      | IL2.TyLit l => IL3.TyLit l
      | IL2.TyArrow(c1, c2) => IL3.TyArrow(erase_con c1, erase_con c2)
      | IL2.TyRef c => IL3.TyRef(erase_con c)
      | IL2.TyTagged => IL3.TyTagged
      | IL2.TyTag c => IL3.TyTag(erase_con c)
      | IL2.TySum cs => IL3.TySum(map erase_con cs)
      | IL2.TyProd cs => IL3.TyProd(map erase_con cs)
      | IL2.TyForall(a, k, c) => IL3.TyForall(a, convert_kind k, erase_con c)

    fun convert_con G c =
      let
        val c' = translate_con c (IL2S.synth_con_kind G c)
        val ctxsubs = convert_ctx G
        val cout =
          foldl (fn ((a, ci), cout) => IL2.subst_con_con ci a cout) c' ctxsubs
      in
        erase_con cout
      end

    fun convert_exp G (e : IL2.exp) =
      case e of
        IL2.ExVar x => IL3.ExVar x
      | IL2.ExLit l => IL3.ExLit l
      | IL2.ExLam (x, c, e) =>
          let
            val (x, e) = IL2.alpha_vary (x, e) IL2.replace_exp_exp
            val G' = Ctx.extend G (Ctx.DecExpCon(x, c))
            val (c1, c2) =
              case IL2Static.synth_exp_con G e of
                IL2.TyArrow(c1, c2) => (c1 ,c2)
              | _ => raise (Convert "Lambda with non-arrow type")
          in
            IL3.exLam (x , convert_con G c1, convert_con G c2, convert_exp G' e)
          end
      | IL2.ExUnOp (oper, e) =>
          IL3.ExUnOp(oper, convert_exp G e)
      | IL2.ExBinOp (e1, oper, e2) =>
          IL3.ExBinOp(convert_exp G e1, oper, convert_exp G e2)
      | IL2.ExFix fbnds =>
        (* we need to convert a function bundle as product to a function
           bundle as projection. We do this by making a pass through 
           function which contains all of the actual components as
           elements of a product and then forcing it. In the case
           where there is only one function we can just dump that
           in a product. *)

        (case fbnds of
          [(f, x, c1, c2, e)] =>
            let
              val (f, e) = IL2.alpha_vary (f, e) IL2.replace_exp_exp
              val (x, e) = IL2.alpha_vary (x, e) IL2.replace_exp_exp
              val G' = Ctx.extend G (Ctx.DecExpCon(f, IL2.TyArrow(c1, c2)))
              val G' = Ctx.extend G' (Ctx.DecExpCon(x, c1))
              val e' = convert_exp G' e
              val func =
                IL3.ExFix(1, [(f, x, convert_con G c1, convert_con G c2, e')])
            in
              IL3.ExProd[func]
            end
        | fbnds =>
            let
              val fbnds = map (fn (f, x, c1, c2, e) =>
                    (f, Var.freshen f, x, Var.freshen x, c1, c2, e)) fbnds
              fun vary_fbnd fbnds (f, g, x, y, c1, c2, e) =
                let
                  val re = IL2.replace_exp_exp
                  val e' =
                    foldl (fn ((f, g, _, _, _, _, _), e) => re f g e) e fbnds
                  val e'' = re x y e'
                in
                  (g, y, c1, c2, e'')
                end
              val fbnds = map (vary_fbnd fbnds) fbnds
              val G' = Ctx.extend_many G
                (map (fn x => Ctx.DecExpCon(#1 x,
                                IL2.TyArrow(#3 x, #4 x))) fbnds)
              fun convert_fbnd (f, x, c1, c2, e) =
                let
                  val G'' = Ctx.extend G' (Ctx.DecExpCon(x, c1))
                  val e' = convert_exp G'' e
                  val c1' = convert_con G'' c1
                  val c2' = convert_con G'' c2
                in
                  (f, x, c1', c2', e')
                end
              val fbnds = map convert_fbnd fbnds
              val len = length fbnds
              val (fnames, fcons)  =
                ListPair.unzip (map (fn (f, _, c1, c2, _) =>
                  (IL3.ExVar f, IL3.TyArrow(c1, c2))) fbnds)
              val f = Var.newNamed "wrapper"
              val x = Var.newNamed "unused"
              val fbody =
                (f, x, IL3.tyUnit, IL3.TyProd fcons, IL3.ExProd fnames)
            in
              IL3.ExApp(IL3.ExFix(1, fbody::fbnds), IL3.exUnit)
            end)
      | IL2.ExApp(e1, e2) => IL3.ExApp(convert_exp G e1, convert_exp G e2)
      | IL2.ExHandle(e1,  e2) =>
          let
            val e1' = convert_exp G e1
            val e2' = convert_exp G e2
            val x = Var.newNamed "x"
          in
            IL3.ExHandle(e1', (x, IL3.ExApp(e2', IL3.ExVar x)))
          end
      | IL2.ExRaise(c, e) => IL3.ExRaise(convert_con G c, convert_exp G e)
      | IL2.ExRef e => IL3.ExRef (convert_exp G e)
      | IL2.ExGet e => IL3.ExGet (convert_exp G e)
      | IL2.ExSet(e1, e2) => IL3.ExSet(convert_exp G e1, convert_exp G e2)
      | IL2.ExRoll(c, e) => IL3.ExRoll(convert_con G c, convert_exp G e)
      | IL2.ExUnroll e => IL3.ExUnroll(convert_exp G e)
      | IL2.ExProd es => IL3.ExProd(map (convert_exp G) es)
      | IL2.ExProj(i, e) => IL3.ExProj(i, convert_exp G e)
      | IL2.ExSumInj(c, i, e) =>
          IL3.ExSumInj(convert_con G c, i, convert_exp G e)
      | IL2.ExCase(c, e, es) =>
          IL3.ExCase(convert_con G c, convert_exp G e, map (convert_exp G) es)
      | IL2.ExTagInj(e1, e2) =>
          IL3.ExTagInj(convert_exp G e1, convert_exp G e2)
      | IL2.ExNewTag c => IL3.ExNewTag(convert_con G c)
      | IL2.ExIfTagOf(e1, e2, e3, e4) =>
          IL3.ExIfTagOf(convert_exp G e1, convert_exp G e2,
                        convert_exp G e3, convert_exp G e4)
      | IL2.ExConAbs(a, k, e) =>
          let
            val (a, e) = IL2.alpha_vary (a, e) IL2.replace_con_exp
            val G' = Ctx.extend G (Ctx.DecConKind(a, k))
          in
            IL3.ExConAbs(a, convert_kind k, convert_exp G' e)
          end
      | IL2.ExConApp(e, c) => IL3.ExConApp(convert_exp G e, convert_con G c)
      | IL2.ExLet(x, e1, e2) =>
        let
          val (x, e2) = IL2.alpha_vary (x, e2) IL2.replace_exp_exp
          val c = IL2S.synth_exp_con G e1
          val G' = Ctx.extend G (Ctx.DecExpCon(x, c))
        in
          IL3.ExLet(x, convert_exp G e1, convert_exp G' e2)
        end
  end
