structure EL =
  struct
    type ident = string
    type tyvar = string

    type longid = ident list * ident
                (* e.g. (["Control", "Print"], "printDepth") *)

    datatype ty =
        TyBool
      | TyUnit
      | TyLit of Literal.lit_ty
      | TyProd of ty * ty
      | TyList of ty
      | TyArrow of ty * ty
      | TyCon of ty list * longid
      | TyVar of string

    datatype exp =
        ExVar of longid
      | ExUnit
      | ExBool of bool
      | ExIf of exp * exp * exp
      | ExLit of Literal.lit
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
      | ExPair of exp * exp
      | ExProj1 of exp
      | ExProj2 of exp
      | ExNil
      | ExCons of exp * exp
      | ExCase of exp * exp * (ident * ident * exp)
      | ExFn of ident * exp
      | ExApp of exp * exp
      | ExLet of binding list * exp
      | ExConstrain of exp * ty

    and binding =
        BndVal of ident * exp
      | BndFun of ident * ident * exp
      | BndType of tyvar list * ident * ty
      | BndOpen of longid
      | BndLocal of binding list * binding list
      | BndStructure of ident * strexp

    and strexp =
        StrVar of longid
      | StrStruct of binding list
      | StrSealOpaque of strexp * sigexp (* strexp :> sigexp *)
      | StrSealTransparent of strexp * sigexp (* strexp : sigexp *)
      | StrLet of binding list * strexp

    and sigexp =
        SigSig of spec list

    and spec =
        SpecVal of ident * ty
      | SpecType of tyvar list * ident * ty option
      | SpecStructure of ident * sigexp

    type program = binding list
  end
