structure LittleFront =
struct

  local
    structure CPSConvert = CPSAdaptorFn(CPSConvert)

    fun compile_program parser pinput=
      let
        val bigelprog = parser pinput
        val elprog = Reparser.reparse_program bigelprog
        val () = print "Elaborating\n"
        val (sbnds, sdecs) = Elaborate.elab_program [] elprog
        val () = print "Expanding Source Language\n"
        val il1prog = IL1.ModStruct (Injector.convert_sbnds [] sbnds)
        val _ = IL1Static.synth_mod_sig IL1Context.empty il1prog
        val () = print "Phase splitting\n"
        val IL2.Mod(c, il2prog) =
          PhaseSplit.split_module il1prog
        val () = print "Eliminating Singleton Kinds\n"
        val il3prog = SingElim.convert_exp IL2Static.Ctx.empty il2prog
        val _ = IL3Static.synth_exp_con IL3Static.Ctx.empty il3prog
        val () = print "CPS Converting\n"
        val il4prog = CPSConvert.convert_program IL3Static.Ctx.empty il3prog
        val () = print "Closure Converting\n"
        val il5prog =
          ClosureConvert.convert_exp IL4Static.Ctx.empty il4prog
        val () = print "Hoisting\n"
        val il6prog = Hoist.convert_exp IL5Static.Ctx.empty il5prog
        val _ = IL6Static.check_program il6prog
        val () = print "Explicating Initialization\n"
        val il7prog = Alloc.convert_program il6prog
        val () = print "Generating Code\n"
        val outprog = GenC.generate_program il7prog
      in
        outprog
      end
      handle IL1Static.Error s =>
          let val () = print ("Got IL1Static Error: "  ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | IL1Static.Equiv =>
          let val () = print ("Got IL1Static Type Equivalence Error\n")
          in OS.Process.exit OS.Process.failure
          end
      | Parser.Parse(p, s) =>
          let val () = print (
                "Parse error: "  ^ s ^ "\n" ^
                "at " ^ (Pos.toString p) ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | Elaborate.Elab  s =>
          let val () = print (
                "Elaboration error: "  ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | IL1Static.Check s =>
          let val () = print ("Got IL1Static Check: "  ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | Elaborate.Error s =>
          let val () = print ("Got Elaboration Error: "  ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | IL4Static.Check s =>
          let val () = print ("Got IL4Static Check: "  ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | ClosureConvert.Error s =>
          let val () = print ("Closure Conversion Error: " ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | IL5Static.Check s =>
          let val () = print ("Got IL5Static Error: "  ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | IL6Static.Error s =>
          let val () = print ("Got IL6Static Error: "  ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
      | IL6Static.Check s =>
          let val () = print ("Got IL6Static Error: "  ^ s ^ "\n")
          in OS.Process.exit OS.Process.failure
          end
  in
    fun compile_file infilename outfilename =
      let
        val outprog = compile_program Parser.parseFile infilename
        val outfile = TextIO.openOut outfilename
        val () = TextIO.output(outfile, outprog)
        val () = TextIO.closeOut outfile
      in
        ()
      end

    fun compile_string instring = compile_program Parser.parseString instring
  end
end
