structure Top : sig
                  val compile_file : string -> string ->
                                     (string -> string -> unit) ->
                                     string * string list -> OS.Process.status
                 end =
struct
  datatype cmd_args = Outfile of string
  val optFields =
    [
      {short = "o",
       long = ["output"],
       desc = GetOpt.ReqArg(fn s => Outfile s, "output file"),
       help = "sets output file"}
    ]
  fun handleOptErr args s =
    let
      val usage = GetOpt.usageInfo
                {header = "hotc 1.0",
                 options = optFields}
    in
      print "Error parsing command line arguments:\n";
      List.app (fn s => print (s ^ " ")) args;
      print "\n";
      print s;
      print "\n";
      print usage;
      print "\n";
      OS.Process.exit (OS.Process.failure)
    end
  fun compile_file rundir gcfile compile_fn (name, args) =
    let
      val opts = GetOpt.getOpt
                    {argOrder = GetOpt.Permute,
                     options = optFields,
                     errFn = handleOptErr args} args
      val (infile, outfile) =
        case opts of
            ([Outfile f], [infile]) => (infile, f)
          | ([], [infile]) => (infile, "a.out")
          | _ => handleOptErr args ""
      val () = compile_fn infile (infile ^ ".c")
      val cwd = OS.FileSys.getDir()
      val runpath = rundir ^ "/runtime.c"
      val gcspath = rundir ^ "/" ^ gcfile
      val buildcommand =
            ("gcc -o " ^ outfile ^ " -g -falign-functions -I" ^ rundir ^ " " ^
              runpath ^ " " ^ gcspath ^ " " ^ (infile ^ ".c"))
      val () = print (buildcommand ^ "\n")
    in
      OS.Process.system buildcommand
    end
end
