(* NOTE: We assume the label list on a module projections
         tuple projection, and sum injection are correct.
         This should probably be fixed in the long run,
         these were hacked in to accomodate a syntactic definition
         of phase splitting *)


structure IL1Static :> IL1STATIC =
  struct
    open IL1
    structure Ctx = IL1Context
    type context = IL1Context.context

    exception Equiv
    exception Occurs
    exception Subkind
    exception Subsig
    exception Check of string
    exception Error of string

    fun check b f = if b then () else f ()

    fun nth (l, n) = SOME (List.nth (l,n)) handle Subscript => NONE


    (*** module values and selfification ***)
    fun module_value module =
        case module of
            ModVar _ => true
          | ModStruct sbnds => sbnds_value sbnds
          | ModAbs _ => true
          | ModApp _ => false
          | ModModProj (m, _, _) => module_value m
          | ModConstrain _ => false
          | ModLet _ => false

    and sbnds_value sbnds =
        case sbnds of
            [] => true
          | sb::sbs => sbnd_value sb andalso sbnds_value sbs

    and sbnd_value (l, bnd) =
        case bnd of
            BndExp (ex, e) => exp_value e
          | BndCon _ => true (* XXX should we check modv.lab cons? *)
          | BndMod (mx, m) => module_value m

    and exp_value exp =
        case exp of
            ExVar ex => true
          | ExLit _ => true
          | ExUnOp _ => false
          | ExBinOp _ => false
          | ExFix _ => true
          | ExApp _ => false
          | ExRecord labexps => List.all (fn (l, e) => exp_value e) labexps
          | ExRecordProj _ => false
          | ExHandle _ => false
          | ExRaise _ => false
          | ExRef _ => false
          | ExGet _ => false
          | ExSet _ => false
          | ExRoll (c, e) => exp_value e
          | ExUnroll e => exp_value e
          | ExSumInj (c, lab, e, labs) => exp_value e
          | ExCase _ => false
          | ExTagInj (e1, e2) => exp_value e1 andalso exp_value e2
          | ExNewTag _ => false
          | ExIfTagOf _ => false
          | ExConAbs (cx, k, e) => exp_value e (* XXX value restriction? *)
          | ExConApp (e, c) => exp_value e
          | ExLet _ =>  false
          | ExModLet _ => false
          | ExModProj (m, l, _) => module_value m

    (* NB: only selfifies structures, since SML functors are generative *)
    fun selfify modv sdecs labs =
        case sdecs of
            [] => []
          | (l, DecExpCon (ex, c)) :: sds =>
                (l, DecExpCon (ex, c))
                :: selfify modv sds labs
          | (l, DecConKind (cx, k)) :: sds =>
                (l, DecConKind (cx, labelledKndSing (ConModProj (modv, l, labs)) k))
                :: selfify modv sds labs
          | (l, DecModSig (mx, SigStruct sds')) :: sds =>
                let val sds' = selfify (ModModProj (modv, l, labs)) sds' (labdom sds')
                in
                    (l, DecModSig (mx, SigStruct sds'))
                    :: selfify modv sds labs
                end
          | (sd as (l, DecModSig (mx, SigPi _))) :: sds =>
                sd :: selfify modv sds labs


    (*** unification proper (used in equivalence below) ***)

    fun unify ebr con =
        case !ebr of
            EvBound _ => raise Error "bound Evar not eliminated by whnf"
          | EvFree (condeps, moddeps, id) =>
                (check_unifiable_con (ConVarSet.empty, ModVarSet.empty)
                                     (condeps, moddeps, id) con;
                 ebr := EvBound con)

    (* checks that
            (a) the Evar with id doesn't occur in con (the "occurs check"), and
            (b) that any free constructor or module variables in con appear in
                condeps or moddeps, respectively.
       if both conditions hold, then `ConEvar (EvFree (condeps, moddeps, id))'
       is unifiable with con.  (keeps track along the way of bound constructor
       variables in bcvs and bound module variables in bmvs.) *)
    and check_unifiable_con (bcvs, bmvs) (condeps, moddeps, id) con =
        (* NB: implemented in terms of IL1.map_*, but never uses the return
           result, so the functions just return dummy values *)
        map_con (fn cx =>
                    if ConVarSet.member (bcvs, cx) orelse
                       ConVarSet.member (condeps, cx)
                    then
                        con
                    else
                        raise Occurs)
                (fn ebr =>
                    case !ebr of
                        EvBound c =>
                            check_unifiable_con (bcvs, bmvs)
                                                (condeps, moddeps, id) c
                      | EvFree (cd, md, id') =>
                            if id = id' then
                              raise Occurs
                            else
                              let val cd' = ConVarSet.intersection (condeps, cd)
                                  val md' = ModVarSet.intersection (moddeps, md)
                              in
                                  ebr := EvFree (cd', md', id');
                                  con
                              end)
                (fn (cx, c) =>
                    (check_unifiable_con (ConVarSet.add (bcvs, cx), bmvs)
                                         (condeps, moddeps, id) c;
                     (cx, c)))
                (check_unifiable_kind (bcvs, bmvs) (condeps, moddeps, id))
                (check_unifiable_module (bcvs, bmvs) (condeps, moddeps, id))
                con

    and check_unifiable_kind (bcvs, bmvs) (condeps, moddeps, id) knd =
        map_kind (fn (cx, k) =>
                    (check_unifiable_kind (ConVarSet.add (bcvs, cx), bmvs)
                                          (condeps, moddeps, id) k;
                     (cx, k)))
                 (check_unifiable_con (bcvs, bmvs) (condeps, moddeps, id))
                 knd

    and check_unifiable_module (bcvs, bmvs) (condeps, moddeps, id) module =
        map_module (fn mx =>
                        if ModVarSet.member (bmvs, mx) orelse
                           ModVarSet.member (moddeps, mx)
                        then
                            module
                        else
                            raise Occurs)
                   (check_unifiable_sbnds (bcvs, bmvs) (condeps, moddeps, id))
                   (fn (mx, m) =>
                        (check_unifiable_module (bcvs, ModVarSet.add (bmvs, mx))
                                                (condeps, moddeps, id) m;
                         (mx, m)))
                   (check_unifiable_signat (bcvs, bmvs) (condeps, moddeps, id))
                   module

    and check_unifiable_sbnds (bcvs, bmvs) (condeps, moddeps, id) sbnds =
        map_sbnds (fn (ex, sbs) =>
                        (check_unifiable_sbnds (bcvs, bmvs)
                                               (condeps, moddeps, id) sbs;
                         (ex, sbs)))
                  (fn (cx, sbs) =>
                        (check_unifiable_sbnds (ConVarSet.add (bcvs, cx), bmvs)
                                               (condeps, moddeps, id) sbs;
                         (cx, sbs)))
                  (fn (mx, sbs) =>
                        (check_unifiable_sbnds (bcvs, ModVarSet.add (bmvs, mx))
                                               (condeps, moddeps, id) sbs;
                         (mx, sbs)))
                  (check_unifiable_exp (bcvs, bmvs) (condeps, moddeps, id))
                  (check_unifiable_con (bcvs, bmvs) (condeps, moddeps, id))
                  (check_unifiable_module (bcvs, bmvs) (condeps, moddeps, id))
                  sbnds

    and check_unifiable_signat (bcvs, bmvs) (condeps, moddeps, id) signat =
        map_signat (fn (mx, sg) =>
                        (check_unifiable_signat (bcvs, ModVarSet.add (bmvs, mx))
                                                (condeps, moddeps, id) sg;
                         (mx, sg)))
                   (check_unifiable_sdecs (bcvs, bmvs) (condeps, moddeps, id))
                   signat

    and check_unifiable_sdecs (bcvs, bmvs) (condeps, moddeps, id) sdecs =
        map_sdecs (fn (ex, sds) =>
                        (check_unifiable_sdecs (bcvs, bmvs)
                                               (condeps, moddeps, id) sds;
                         (ex, sds)))
                  (fn (cx, sds) =>
                        (check_unifiable_sdecs (ConVarSet.add (bcvs, cx), bmvs)
                                               (condeps, moddeps, id) sds;
                         (cx, sds)))
                  (fn (mx, sds) =>
                        (check_unifiable_sdecs (bcvs, ModVarSet.add (bmvs, mx))
                                               (condeps, moddeps, id) sds;
                         (mx, sds)))
                  (check_unifiable_con (bcvs, bmvs) (condeps, moddeps, id))
                  (check_unifiable_kind (bcvs, bmvs) (condeps, moddeps, id))
                  (check_unifiable_signat (bcvs, bmvs) (condeps, moddeps, id))
                  sdecs

    and check_unifiable_exp (bcvs, bmvs) (condeps, moddeps, id) exp =
        map_exp (fn ex => exp) (* no checking to be done for a expvar *)
                (fn fbnds =>
                    map (fn (f, x, c1, c2, e) =>
                            (f, x, check_unifiable_con (bcvs, bmvs)
                                               (condeps, moddeps, id) c1,
                                   check_unifiable_con (bcvs, bmvs)
                                               (condeps, moddeps, id) c2,
                                   check_unifiable_exp (bcvs, bmvs)
                                               (condeps, moddeps, id) e))
                        fbnds)
                (fn (ex, e) =>
                    (check_unifiable_exp (bcvs, bmvs) (condeps, moddeps, id) e;
                     (ex, e)))
                (fn (cx, e) =>
                    (check_unifiable_exp (ConVarSet.add (bcvs, cx), bmvs)
                                         (condeps, moddeps, id) e;
                     (cx, e)))
                (fn (mx, e) =>
                    (check_unifiable_exp (bcvs, ModVarSet.add (bmvs, mx))
                                         (condeps, moddeps, id) e;
                     (mx, e)))
                (check_unifiable_con (bcvs, bmvs) (condeps, moddeps, id))
                (check_unifiable_kind (bcvs, bmvs) (condeps, moddeps, id))
                (check_unifiable_module (bcvs, bmvs) (condeps, moddeps, id))
                exp


    (*** static semantics ***)

    (* precondition: module is a path *)
    fun extract_natural_sig ctx module =
        case module of
            ModVar mx =>
                (case Ctx.lookup_modvar ctx mx of
                    NONE => raise Check
                            ("extract_natural_sig: unbound modvar, "
                            ^ Variable.toString mx)
                  | SOME sg => sg)
          | ModModProj (modv, lab, labs) =>
                let val sdecs =
                        case extract_natural_sig ctx modv of
                            SigStruct sds => sds
                          | _ => raise Error ("extract_natural_sig: "
                                ^ "module projection from non-structure")
                in
                    case Ctx.sdecs_lookup_modlab modv sdecs lab of
                        SOME sg => sg
                      (* | SOME _ => raise Error "extract_natural_sig: bad sig" *)
                      | NONE =>
                            raise Error
                            ("extract_natural_sig: no such module label, "
                            ^ string_of_label lab)
                end
          | _ => raise Error "extract_natural_sig: applied to a non-path"

    (* precondition: con is a path *)
    fun extract_natural_kind ctx con =
        case con of
            ConVar cx =>
                (case Ctx.lookup_convar ctx cx of
                    NONE => raise Check
                            ("extract_natural_kind: unbound convar, "
                            ^ Variable.toString cx)
                  | SOME k => k)
          | ConApp (p, c) =>
                (case extract_natural_kind ctx p of
                    KndPi (cx, k, l) => subst_con_kind c cx l
                  | _ => raise Error
                         "extract_natural_kind: function not of Pi kind")
          | ConProj (i, p) =>
                (case extract_natural_kind ctx p of
                    KndProd ks =>
                        (case nth (ks, i-1) of
                            SOME k => k
                          | NONE =>
                                raise Error
                                ("extract_natural_kind: subscript out of bounds, "
                                ^ Int.toString i))
                  | _ => raise Error
                         "extract_natural_kind: projection from non-product")
          | ConMu (cx, k, c) => k
          | ConModProj (modv, lab, labs) =>
                (case extract_natural_sig ctx modv of
                    SigStruct sdecs =>
                        (case Ctx.sdecs_lookup_conlab modv sdecs lab of
                            SOME k => k
                          | NONE =>
                                raise Error
                                ("extract_natural_kind: no such module label, "
                                ^ string_of_label lab))
                  | _ => raise Error
                         ("extract_natural_kind: "
                         ^ "module projection from non-structure"))
          | _ => raise Error "extract_natural_kind: applied to a non-path"

    fun whnf ctx con =
        let fun reduce_mod ctx module =
                case module of
                    ModModProj (modv as ModStruct sbnds, lab, labs) =>
                        (case Ctx.sbnds_lookup_modlab modv sbnds lab of
                            SOME m => SOME m
                          | NONE => raise Error
                                    ("reduce_mod: no such module label, "
                                    ^ string_of_label lab))
                  | ModModProj (modv, lab, labs) =>
                        (case reduce_mod ctx modv of
                            SOME modv' => SOME (ModModProj (modv', lab, labs))
                          | NONE => NONE)
                  | _ => NONE
            fun path_reduce ctx conpath =
                case extract_natural_kind ctx conpath of
                    KndSing c => SOME c
                  | _ => NONE
            fun reduce ctx con =
                case con of
                  (* beta reductions *)
                    ConApp (ConAbs (cx, k, c1), c2) =>
                        SOME (subst_con_con c2 cx c1)
                  | ConProj (i, ConProd cs) =>
                        (case nth (cs, i - 1) of
                            SOME c => SOME c
                          | NONE => raise Error
                                    ("reduce: subscript out of bounds, "
                                    ^ Int.toString i))
                  | ConModProj (modv as ModStruct sbnds, lab, labs) =>
                        (case Ctx.sbnds_lookup_conlab modv sbnds lab of
                            SOME c => SOME c
                          | NONE => raise Error("reduce: no such module label, "
                                                ^ string_of_label lab))
                  (* congruences *)
                  | ConApp (c1, c2) =>
                        (case reduce ctx c1 of
                            SOME c1' => SOME (ConApp (c1', c2))
                          | NONE => path_reduce ctx con)
                  | ConProj (i, c) =>
                        (case reduce ctx c of
                            SOME c' => SOME (ConProj (i, c'))
                          | NONE => path_reduce ctx con)
                  | ConModProj (modv, lab, labs) =>
                        (case reduce_mod ctx modv of
                            SOME modv' => SOME (ConModProj (modv', lab, labs))
                          | NONE => path_reduce ctx con)
                  (* definition resolution *)
                  | ConVar _ => path_reduce ctx con
                  (* robustness wrt bound evars *)
                  | ConEvar ebr =>
                        (case !ebr of
                            EvBound c => SOME c
                          | EvFree _ => NONE)
                  | _ => NONE
        in
            case reduce ctx con of
                SOME con' => whnf ctx con'
              | NONE => con
        end


    fun subkind ctx knd1 knd2 =
        case (knd1, knd2) of
            (KndType, KndType) => ()
          | (KndSing c, KndType) => ()
          | (KndSing c1, KndSing c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                             handle Equiv => raise Subkind
                in
                    ()
                end
          | (KndPi (cx1, k1, l1), KndPi (cx2, k2, l2)) =>
                let val () = subkind ctx k2 k1
                    val (cx, (l1, l2)) = alpha_vary2 (cx1, l1) (cx2, l2)
                                                     replace_con_kind
                    val ctx = Ctx.extend ctx (DecConKind (cx, k2))
                    val () = subkind ctx l1 l2
                in
                    ()
                end
          | (KndProd knds1, KndProd knds2) =>
                let val () = check (List.length knds1 = List.length knds2)
                             (fn () => raise Subkind)
                    val () =
                        ListPair.app
                            (fn (k1, k2) => subkind ctx k1 k2)
                            (knds1, knds2)
                in
                    ()
                end
          | _ => raise Subkind

    and subsig ctx signat1 signat2 =
        case (signat1, signat2) of
            (SigStruct sdecs1, SigStruct sdecs2) => subsdecs ctx sdecs1 sdecs2
          | (SigPi (mx1, sg1, sg1'), SigPi (mx2, sg2, sg2')) =>
                let val () = subsig ctx sg2 sg1
                    val (mx, (sg1', sg2')) = alpha_vary2 (mx1, sg1') (mx2, sg2')
                                                         replace_mod_signat
                    val ctx = Ctx.extend ctx (DecModSig (mx, sg2))
                    val () = subsig ctx sg1' sg2'
                in
                    ()
                end
          | _ => raise Subsig

    and subsdecs ctx sdecs1 sdecs2 =
        case (sdecs1, sdecs2) of
            ([], []) => ()
          | ((lab1, d1)::sds1, (lab2, d2)::sds2) =>
                let val () = check (lab1 = lab2) (fn () => raise Subsig)
                    val (d1, (sds1, sds2)) = subdec ctx (d1, sds1) (d2, sds2)
                    val () = subsdecs (Ctx.extend ctx d1) sds1 sds2
                in
                    ()
                end
          | _ => raise Subsig

    (* does two things in parallel:
        (1) make sure dec1 <: dec2
        (2) alpha_vary2 (dec1, sdecs1) (dec2, sdecs2)
       then returns (dec1', (sdecs1, sdecs2)), where dec1' is the
       alpha-variant of dec1 *)
    and subdec ctx (dec1, sdecs1) (dec2, sdecs2) =
        case (dec1, dec2) of
            (DecExpCon (ex1, c1), DecExpCon (ex2, c2)) =>
                let val (ex, (sdecs1, sdecs2)) = alpha_vary2 (ex1, sdecs1)
                                                             (ex2, sdecs2)
                                                             replace_exp_sdecs
                    val () = equiv_con ctx c1 c2 KndType
                             handle Equiv =>
                                (print "[---\n";
                                 Layout.print
                                  (IL1Print.ctx_to_layout (Ctx.to_decs ctx),
                                        print);
                                 print "\n|-\n";
                                 Layout.print (IL1Print.con_to_layout c1,
                                        print);
                                 print "\n<=>?\n";
                                 Layout.print (IL1Print.con_to_layout c2,
                                        print);
                                 print "\n";
                                 print "---]\n";
                                 raise Match)
                in
                    (DecExpCon (ex, c1), (sdecs1, sdecs2))
                end
          | (DecConKind (cx1, k1), DecConKind (cx2, k2)) =>
                let val (cx, (sdecs1, sdecs2)) = alpha_vary2 (cx1, sdecs1)
                                                             (cx2, sdecs2)
                                                             replace_con_sdecs
                    val () = subkind ctx k1 k2
                             handle Subkind => raise Subsig
                in
                    (DecConKind (cx, k1), (sdecs1, sdecs2))
                end
          | (DecModSig (mx1, sg1), DecModSig (mx2, sg2)) =>
                let val (mx, (sdecs1, sdecs2)) = alpha_vary2 (mx1, sdecs1)
                                                             (mx2, sdecs2)
                                                             replace_mod_sdecs
                    val () = subsig ctx sg1 sg2
                in
                    (DecModSig (mx, sg1), (sdecs1, sdecs2))
                end
          | _ => raise Subsig

    and equiv_con ctx con1 con2 knd =
        case knd of
            KndType =>
                let val k = equiv_con_path ctx (whnf ctx con1) (whnf ctx con2)
                    (* is this kind check necessary?  maybe for the sake of
                       belt and suspenders... *)
                    val () = subkind ctx k KndType
                             handle Equiv => raise Error
                             ("mismatch between equiv_con input (KndType) "
                             ^ " and equiv_con_path output")
                in
                    ()
                end
          | KndSing _ => ()
          | KndProd knds =>
                ListUtil.appi
                        (fn (k, i) =>
                            equiv_con ctx (ConProj (i+1, con1))
                                          (ConProj (i+1, con2))
                                          k)
                        knds
          | KndPi (x, k, l) =>
                let val (x, l) = alpha_vary (x, l) replace_con_kind
                in
                    equiv_con (Ctx.extend ctx (DecConKind (x, k)))
                              (ConApp (con1, ConVar x))
                              (ConApp (con2, ConVar x))
                              l
                end

    and equiv_module_path ctx modp1 modp2 =
        case (modp1, modp2) of
            (ModVar mx1, ModVar mx2) =>
                if Variable.eq (mx1, mx2) then
                    case Ctx.lookup_modvar ctx mx1 of
                        NONE => raise Error
                                ("comparing unbound module variable, "
                                ^ Variable.toString mx1)
                      | SOME sg => sg
                else
                    raise Equiv
          | (ModModProj (mp1, lab1, labs1), ModModProj (mp2, lab2, labs2)) =>
                let val () = check (label_eq (lab1, lab2))
                                   (fn () => raise Equiv)
                    val sdecs =
                            case equiv_module_path ctx mp1 mp2 of
                                SigStruct sds => sds
                              | _ => raise Error
                                    ("equiv_module_path: "
                                    ^ "module projection from non-structure")
                in
                    case Ctx.sdecs_lookup_modlab mp1 sdecs lab1 of
                        NONE => raise Error
                               ("equiv_module_path: non-existent module label, "
                               ^ string_of_label lab1)
                      | SOME sg => sg
                end
          | _ => raise Equiv

    and equiv_con_path ctx con1 con2 =
        case (con1, con2) of
            (ConVar cx1, ConVar cx2) =>
                if Variable.eq (cx1, cx2) then
                    case Ctx.lookup_convar ctx cx1 of
                        NONE => raise Error ("comparing unbound variable, "
                                            ^ Variable.toString cx1)
                      | SOME k => k
                else
                    raise Equiv
          | (ConEvar ebr1, c2 as ConEvar ebr2) =>
                if ebr1 = ebr2 then
                    KndType
                else
                    let val () = unify ebr1 c2
                    in
                        KndType
                    end
          | (ConEvar ebr, c) =>
                let val () = check_con_kind ctx c KndType
                             "filling in evar with non-type"
                    val () = unify ebr c
                in  
                    KndType
                end
          | (c, ConEvar ebr) => 
                let val () = check_con_kind ctx c KndType
                             "filling in evar with non-type"
                    val () = unify ebr c
                in  
                    KndType
                end
          | (ConApp (cp1, c1), ConApp (cp2, c2)) =>
                let val k = equiv_con_path ctx cp1 cp2
                in
                    case k of
                        KndPi (x, k1, k2) =>
                            let val () = equiv_con ctx c1 c2 k1
                            in
                                IL1.subst_con_kind c1 x k2
                            end
                      | _ => raise Error
                             "equiv_con_path: application of non-function"
                end
          | (ConProj (i1, cp1), ConProj (i2, cp2)) =>
                let val () = check (i1 = i2) (fn () => raise Equiv)
                in
                    case equiv_con_path ctx cp1 cp2 of
                        KndProd knds =>
                           (case nth (knds, i1 - 1) of
                               SOME k => k
                             | NONE =>
                                raise Error
                                ("equiv_con_path: subscript out of bounds, "
                                ^ Int.toString i1))
                      | _ => raise Error
                             "equiv_con_path: projection from a non-tuple"
                end
          | (ConMu (cx1, k1, c1), ConMu (cx2, k2, c2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2)
                                                     replace_con_con
                    val () = equiv_con (Ctx.extend ctx (DecConKind (cx,k1)))
                                       c1 c2 k1
                in
                    k1
                end
          | (ConModProj (modp1, lab1, labs1), ConModProj (modp2, lab2, labs2)) =>
                let val () = check (label_eq (lab1, lab2))
                                   (fn () => raise Equiv)
                    val sdecs =
                            case equiv_module_path ctx modp1 modp2 of
                                SigStruct sds => sds
                              | _ => raise Error ("equiv_con_path: "
                                     ^ "module projection from non-structure")
                in
                    case Ctx.sdecs_lookup_conlab modp1 sdecs lab1 of
                        NONE => raise Error
                                ("equiv_con_path: non-existent module label, "
                                ^ string_of_label lab1)
                      | SOME k => k
                end

          | (TyLit lt1, TyLit lt2) =>
                let val () = check (Literal.lit_ty_eq (lt1, lt2))
                                   (fn () => raise Equiv)
                in
                    KndType
                end
          | (TyArrow (c1dom, c1cod), TyArrow (c2dom, c2cod)) =>
                let val () = equiv_con ctx c1dom c2dom KndType
                    val () = equiv_con ctx c1cod c2cod KndType
                in
                    KndType
                end
          | (TyRef c1, TyRef c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                in
                    KndType
                end
          | (TyTagged, TyTagged) => KndType
          | (TyTag c1, TyTag c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                in
                    KndType
                end
          | (TySum labcs1, TySum labcs2) =>
                let val () = check (List.length labcs1 = List.length labcs2)
                                   (fn () => raise Equiv)
                    val () = ListPair.app
                                (fn ((lab1, c1), (lab2, c2)) =>
                                    let val () = check (label_eq (lab1, lab2))
                                                       (fn () => raise Equiv)
                                        val () = equiv_con ctx c1 c2 KndType
                                    in
                                        ()
                                    end)
                                (labcs1, labcs2)
                in
                    KndType
                end
          | (TyRecord labcs1, TyRecord labcs2) =>
                let val () = check (List.length labcs1 = List.length labcs2)
                                   (fn () => raise Equiv)
                    val () = ListPair.app
                                (fn ((lab1, c1), (lab2, c2)) =>
                                    let val () = check (label_eq (lab1, lab2))
                                                       (fn () => raise Equiv)
                                        val () = equiv_con ctx c1 c2 KndType
                                    in
                                        ()
                                    end)
                                (labcs1, labcs2)
                in
                    KndType
                end
          | (TyForall (cx1, k1, c1), TyForall (cx2, k2, c2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2)
                                                     replace_con_con
                    val () = equiv_con (Ctx.extend ctx (DecConKind (cx,k1)))
                                       c1 c2 KndType
                 in
                     KndType
                 end

          | (ConAbs _, ConAbs _) => raise Error "equiv_con_path on non-path"
          | (ConProd _, ConProd _) => raise Error "equiv_con_path on non-path"
          | _ => raise Equiv


    and check_equiv_con ctx con1 con2 knd s =
        equiv_con ctx con1 con2 knd
        handle Equiv => raise Check s

    and equiv_kind ctx knd1 knd2 =
        case (knd1, knd2) of
            (KndType, KndType) => ()
          | (KndPi (cx1, k1, l1), KndPi (cx2, k2, l2)) =>
                let val () = equiv_kind ctx k1 k2
                    val (cx, (l1, l2)) = alpha_vary2 (cx1, l1) (cx2, l2)
                                                     replace_con_kind
                    val ctx = Ctx.extend ctx (DecConKind (cx, k1))
                    val () = equiv_kind ctx l1 l2
                in
                    ()
                end
          | (KndProd knds1, KndProd knds2) =>
                let val () = check (List.length knds1 = List.length knds2)
                                   (fn () => raise Equiv)
                    val () = ListPair.app (fn (k1, k2) => equiv_kind ctx k1 k2)
                                          (knds1, knds2)
                in
                    ()
                end
          | (KndSing c1, KndSing c2) => equiv_con ctx c1 c2 KndType
          | _ => raise Equiv

    (* XXX error-handling could probably be better on these check_* functions,
       but since they're mostly belt-and-suspenders checks, i suppose it's okay
       to keep them as they are for now... *)

    and check_kind ctx knd =
        case knd of
            KndType => ()
          | KndPi (cx, k, l) =>
                let val () = check_kind ctx k
                    val (cx, l) = alpha_vary (cx, l) replace_con_kind
                    val ctx = Ctx.extend ctx (DecConKind (cx, k))
                    val () = check_kind ctx l
                in
                    ()
                end
          | KndProd ks => List.app (fn k => check_kind ctx k) ks
          | KndSing c => check_con_kind ctx c KndType
                         "singleton at kind other than Type"

    and check_sig ctx signat =
        case signat of
            SigStruct sdecs => check_sdecs ctx sdecs
          | SigPi (mx, sg, sg') =>
                let val () = check_sig ctx sg
                    val (mx, sg') = alpha_vary (mx, sg') replace_mod_signat
                    val ctx = Ctx.extend ctx (DecModSig (mx, sg))
                    val () = check_sig ctx sg'
                in
                    ()
                end

    and check_sdecs ctx sdecs =
        case sdecs of
            [] => ()
          | (lab, d)::sds =>
                let val () = check (not (List.exists (fn l => label_eq (l, lab))
                                                     (labdom sds)))
                                   (fn () => raise Check ("check_sdecs: "
                                                ^ "invalid sdecs: label "
                                                ^ string_of_label lab
                                                ^ " bound multiple times"))
                    val () = check_dec ctx d
                    val (d, sds) = alpha_vary_dec_sdecs (d, sds)
                    val ctx = Ctx.extend ctx d
                    val () = check_sdecs ctx sds
                in
                    ()
                end

    and check_dec ctx dec =
        case dec of
            DecExpCon (x, c) =>
                let val () = check_con_kind ctx c KndType
                             "check_dec: invalid DecExpCon: con isn't a type"
                in
                    ()
                end
          | DecConKind (cx, k) =>
                let val () = check_kind ctx k
                in
                    ()
                end
          | DecModSig (mx, sg) =>
                let val () = check_sig ctx sg
                in
                    ()
                end

    and synth_mod_sig ctx module =
        case module of
            ModVar mx =>
                (case Ctx.lookup_modvar ctx mx of
                    SOME (SigStruct sdecs) => SigStruct (selfify module sdecs (labdom sdecs))
                  | SOME (sg as SigPi _) => sg (* no selfification for
                                                  generative functors *)
                  | NONE => raise Check ("unbound modvar, " ^ Variable.toString mx))
          | ModStruct sbnds => SigStruct (synth_sbnds_sdecs ctx sbnds)
          | ModAbs (mx, sg, m) =>
                (* XXX does a ModAbs need to be ascribed with its result sig
                   to ensure principal signatures?  Dreyer does that... *)
                let val () = check_sig ctx sg
                    val (mx, m) = alpha_vary (mx, m) replace_mod_module
                    val ctx = Ctx.extend ctx (DecModSig (mx, sg))
                    val sg' = synth_mod_sig ctx m
                in
                    SigPi (mx, sg, sg')
                end
          | ModApp (m, modv) =>
                (case synth_mod_sig ctx m of
                    SigPi (mx, sg, sg') =>
                        let val () = check (module_value modv)
                                     (fn () => raise Error
                                     "argument in ModApp not a module value")
                            val () = check_mod_sig ctx modv sg
                                     "argument in ModApp has wrong sig"
                        in
                            subst_mod_signat modv mx sg'
                        end
                  | _ => raise Check
                         "synth_mod_sig: application of a non-functor")
          | ModModProj (modv, lab, labs) =>
                let val sdecs =
                        case synth_mod_sig ctx modv of
                            SigStruct sds => sds
                          | _ => raise Check ("synth_mod_sig: "
                                 ^ "module projection from non-structure")
                    (* belt and suspenders check *)
                    val () = if module_value modv then
                                ()
                             else
                                raise Error "projection from a non-value module"
                    val sdecs = selfify modv sdecs labs
                in
                    case Ctx.sdecs_lookup_modlab modv sdecs lab of
                        NONE => raise Check ("no such module label, "
                                            ^ string_of_label lab)
                      | SOME sg => sg
                end
          | ModConstrain (m, sg) =>
                let val () = check_mod_sig ctx m sg
                             "ascribed module doesn't match ascribed signature"
                in
                    sg
                end
          | ModLet (mx, m, m', sg') =>
                let val sg = synth_mod_sig ctx m
                    val () = check_sig ctx sg'
                    val (mx, m') = alpha_vary (mx, m') replace_mod_module
                    val ctx = Ctx.extend ctx (DecModSig (mx, sg))
                    val () = check_mod_sig ctx m' sg'
                             "synth_mod_sig: ModLet body signature incorrect"
                in
                    sg'
                end

    and check_mod_sig ctx module signat s =
        let val signat' = synth_mod_sig ctx module
        in
            subsig ctx signat' signat
            handle Subsig => raise Check s
        end

    and synth_sbnds_sdecs ctx sbnds =
        case sbnds of
            [] => []
          | (lab, bnd)::sbs =>
                let val (bnd, sbs) = alpha_vary_bnd_sbnds (bnd, sbs)
                    val dec = 
                    case bnd of
                        BndExp (ex, e) => DecExpCon (ex, synth_exp_con ctx e)
                      | BndCon (cx, c) => DecConKind (cx, synth_con_kind ctx c)
                      | BndMod (mx, m) => DecModSig (mx, synth_mod_sig ctx m)
                in
                    (lab, dec) :: synth_sbnds_sdecs (Ctx.extend ctx dec) sbs
                end

    and synth_con_kind ctx con =
        case con of
            ConVar cx => 
                (case Ctx.lookup_convar ctx cx of
                    SOME k => labelledKndSing con k
                  | NONE => raise Check ("synth_con_kind: "
                            ^ "unbound convar, " ^ Variable.toString cx))
          | ConEvar ebr =>
                (case !ebr of
                    EvBound c => synth_con_kind ctx c
                  | EvFree _ => KndSing con)
          | ConAbs (cx, k, c) =>
                let val () = check_kind ctx k
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (DecConKind (cx, k))
                    val k' = synth_con_kind ctx c
                in
                    KndPi (cx, k, k')
                end
          | ConApp (c1, c2) =>
                let val k1 = synth_con_kind ctx c1
                in
                    case k1 of
                        KndPi (cx, k, k') =>
                            let val () = check_con_kind ctx c2 k
                                         ("synth_con_kind: "
                                         ^ "argument in ConApp has wrong kind")
                            in
                                subst_con_kind c2 cx k'
                            end
                      | _ => raise Check
                             "synth_con_kind: function in ConApp has wrong kind"
                end
          | ConProd cons => KndProd (map (fn c => synth_con_kind ctx c) cons)
          | ConProj (i, c) =>
                let val k = synth_con_kind ctx c
                in
                    case k of
                        KndProd ks =>
                            (case nth (ks, i - 1) of
                                SOME k => k
                              | NONE => raise Check
                                    ("synth_con_kind: subscript out of bounds, "
                                    ^ Int.toString i))
                      | _ => raise Check
                             ("synth_con_kind: projection from a non-tuple con")
                end
          | ConMu (cx, k, c) =>
                let fun is_Tn k =
                        case k of
                            KndProd ks =>
                                ((List.all
                                    (fn k =>
                                         (* equiv, not subkind -- ConMu at
                                            singletons -- crazy! *)
                                         let val () = equiv_kind ctx k KndType
                                         in  true
                                         end)
                                    ks)
                                 handle Equiv => false)
                          | _ => false
                    val () = check_kind ctx k
                    val () = check (is_Tn k) (fn () => raise Check
                                     "synth_con_kind: ConMu over non-T^n kind")
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (DecConKind (cx, k))
                    val () = check_con_kind ctx c k
                             "synth_con_kind: ConMu body has wrong kind"
                in
                    labelledKndSing con k
                end
          | ConModProj (modv, lab, labs) =>
                let val sdecs =
                        case synth_mod_sig ctx modv of
                            SigStruct sds => sds
                          | _ => raise Check ("synth_con_kind: "
                                 ^ "module projection from a non-structure")
                    (* belt and suspenders check *)
                    val () = if module_value modv then
                                ()
                             else
                                raise Error "projection from a non-value module"
                    val sdecs = selfify modv sdecs labs
                in
                    case Ctx.sdecs_lookup_conlab modv sdecs lab of
                        NONE => raise Check ("no such module label, "
                                            ^ string_of_label lab)
                      | SOME k => k
                end
          | TyLit _ => KndSing con
          | TyArrow (c1, c2) =>
                let val () = check_con_kind ctx c1 KndType 
                            "synth_con_kind: TyArrow with non-type component"
                    val () = check_con_kind ctx c2 KndType
                            "synth_con_kind: TyArrow with non-type component"
                in
                    KndSing con
                end
          | TyRef c =>
                let val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyRef with non-type argument"
                in
                    KndSing con
                end
          | TyTagged => KndSing con
          | TyTag c =>
                let val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyTag with non-type argument"
                in
                    KndSing con
                end
          | TySum labcs =>
                let val () = List.app
                                (fn (l, c) => check_con_kind ctx c KndType
                                              ("synth_con_kind: TySum "
                                              ^ "with non-type component"))
                                labcs
                in
                    KndSing con
                end
          | TyRecord labcs =>
                let val () = List.app
                                (fn (l, c) => check_con_kind ctx c KndType
                                              ("synth_con_kind: TyRecord "
                                              ^ "with non-type component"))
                                labcs
                in
                    KndSing con
                end
          | TyForall (cx, k, c) =>
                let val () = check_kind ctx k
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (DecConKind (cx, k))
                    val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyForall with non-type body"
                in
                    KndSing con
                end

    and check_con_kind ctx con knd s =
        let val knd' = synth_con_kind ctx con
        in
            subkind ctx knd' knd
            handle Subkind => raise Check s
        end

    (* expansion for recursive kinds, a la Dreyer *)
    and expand ctx recpath =
        let (* decomposes a path into its elimination context and its root
               if [ decompose_path path ] = (elim_ctx, root),
               then [ elim_ctx root ] = path *)
            fun decompose_path path =
                case path of
                    ConProj (i, cp) =>
                        let val (elim_ctx, root) = decompose_path cp
                        in
                            (fn c => ConProj (i, elim_ctx c), root)
                        end
                  | ConApp (cp, c_arg) =>
                        let val (elim_ctx, root) = decompose_path cp
                        in
                            (fn c => ConApp (elim_ctx c, c_arg), root)
                        end
                  | _ => (fn c => c, path)
            val () = check_con_kind ctx recpath KndType
                     "expand: recursive path is not a type"
            val () = equiv_kind ctx (extract_natural_kind ctx recpath) KndType
                     handle Equiv => raise Check
                     "expand: recursive path's natural kind not KndType"
            val (elim_ctx, c_mu) = decompose_path recpath
        in
            case c_mu of
                ConMu (cx, k, c) => elim_ctx (subst_con_con c_mu cx c)
              | _ => raise Check "expand: recursive path not rooted at ConMu"
        end

    and synth_exp_con ctx exp =
        case exp of
            ExVar x => 
                (case Ctx.lookup_expvar ctx x of
                    SOME ty => ty
                  | NONE =>
                      raise Check ("unbound expvar, " ^ Variable.toString x))
          | ExLit l => TyLit (Literal.lit_ty l)
          | ExUnOp(oper, e) =>
                let val (cin, cout) = Literal.unop_tys oper TyLit tyBool tyUnit
                    val () = check_exp_con ctx e cin
                             "synth_exp_con: ExUnOp argument incorrect"
                in
                  cout
                end
          | ExBinOp(e1, oper, e2) =>
                let val (cin1, cin2, cout) =
                          Literal.binop_tys oper TyLit tyBool tyUnit
                    val () = check_exp_con ctx e1 cin1
                             "synth_exp_con: ExBinOp first argument incorrect"
                    val () = check_exp_con ctx e2 cin2
                             "synth_exp_con: ExBinOp second argument incorrect"
                in
                    cout
                end
          | ExFix fbnds =>
                let (* every function body is checked in an enriched scope
                       including bindings for all the functions *)
                       (* XXX should technically be doing some massive
                          alpha-varying, but this should work since no
                          classifier can depend on expression variables *)
                    val ctx0 = ctx (* save original context *)
                    val ctx = List.foldl
                                (fn ((f, _, c, c', _), ctx) =>
                                      Ctx.extend ctx
                                      (DecExpCon (f, TyArrow (c, c'))))
                                ctx
                                fbnds
                    (* each individual function body is checked in a scope
                       enriched with the type of its formal parameter *)
                    fun check_fbnd ctx (f, x, c, cbody, ebody) =
                        let val ctx = Ctx.extend ctx (DecExpCon (x, c))
                            val () = check_con_kind ctx0 c KndType
                                     "synth_exp_con: ExFix domain not a type"
                            val () = check_con_kind ctx0 cbody KndType
                                     "synth_exp_con: ExFix co-domain not a type"
                            val () = check_exp_con ctx ebody cbody
                                   ("synth_exp_con: function body has wrong" ^
                                    "type in function " ^ (Variable.toString f))
                        in
                            ()
                        end
                    val () = List.app (check_fbnd ctx) fbnds
                in
                    (* the resulting type is a product of functions *)
                    tyProd (List.map (fn (_, _, c, c', _) => TyArrow (c, c'))
                                     fbnds)
                end
          | ExApp (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf ctx c1 of
                        TyArrow (c1dom, c1cod) =>
                            (* XXX use check_exp_con instead? *)
                            let val c2 = synth_exp_con ctx e2
                                val () = check_equiv_con ctx c1dom c2 KndType
                                         ("synth_exp_con: function argument "
                                         ^ "doesn't match function domain")
                            in
                                c1cod
                            end
                      | _ => raise Check
                             "synth_exp_con: application of a non-function"
                end
          | ExRecord labes =>
                (* check uniqueness of labels? *)
                TyRecord (map (fn (lab, e) => (lab, synth_exp_con ctx e)) labes)
          | ExRecordProj (lab, e, labs) =>
                let val c = synth_exp_con ctx e
                in
                    case whnf ctx c of
                        TyRecord labtys =>
                            (case ListUtil.Alist.find label_eq labtys lab of
                                NONE => raise Check ("synth_exp_con: "
                                        ^ "no such record label, "
                                        ^ string_of_label lab)
                              | SOME ty => ty)
                      | _ => raise Check
                             "synth_exp_con: projection from a non-record"
                end
          | ExHandle (e1, e2) =>
                (* XXX could be simpler, at a risk of worse error reporting,
                   by just using check_exp_con on e2 *)
                let val c1 = synth_exp_con ctx e1
                    val c2 = synth_exp_con ctx e2
                in
                    case whnf ctx c2 of
                        TyArrow (tytagged, c1') =>
                            (case whnf ctx tytagged of
                                TyTagged =>
                                    let val () = check_equiv_con ctx c1 c1'
                                                 KndType
                                                 ("exception handler has "
                                                 ^ "wrong return type")
                                    in
                                        c1
                                    end
                              | _ => raise Check ("synth_exp_con: exception "
                                     ^ "handler domain type not TyTagged"))
                      | _ => raise Check ("synth_exp_con: "
                             ^ "exception handler not of arrow type")
                end
          | ExRaise (c, e) =>
                let val ce = synth_exp_con ctx e
                    val $ = Layout.str
                    val () = check_equiv_con ctx ce TyTagged KndType
                             (Layout.tostring
                             (Layout.align
                               [$"synth_exp_con: raise arg not of type tagged",
                                $"found type:",
                                IL1Print.con_to_layout ce,
                                $"in expresssion:",
                                IL1Print.exp_to_layout e
                               ]))
                    val () = check_con_kind ctx c KndType
                             "synth_exp_con: con arg to raise not a type"
                in
                    c
                end
          | ExRef e => TyRef(synth_exp_con ctx e)
          | ExGet e =>
                let val c = synth_exp_con ctx e
                in
                    case whnf ctx c of
                        TyRef c' => c'
                      | _ => raise Check
                             "synth_exp_con: ExGet applied to a non-ref"
                end
          | ExSet (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf ctx c1 of
                        TyRef c2 =>
                            let val () = check_exp_con ctx e2 c2
                                         ("synth_exp_con: ExSet arg has "
                                         ^ "wrong type")
                            in
                                tyUnit
                            end
                      | _ => raise Check
                             "synth_exp_con: ExSet applied to a non-ref"
                end
          (* see Dreyer's dissertation for the Roll/Unroll rules *)
          | ExRoll (c, e) =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExRoll at non-type"
                    val recpath = whnf ctx c
                    val c_expanded = expand ctx recpath
                    val () = check_exp_con ctx e c_expanded
                             "synth_exp_con: ExRoll arg has wrong type"
                in
                    recpath
                end
          | ExUnroll e =>
                let val c = synth_exp_con ctx e
                    val recpath = whnf ctx c
                    val c_expanded = expand ctx recpath
                in
                    c_expanded
                end
          | ExSumInj (c, lab, e, labs) =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExSumInj at non-type"
                    val c = whnf ctx c
                in
                    case c of
                        TySum labtys =>
                            (case ListUtil.Alist.find label_eq labtys lab of 
                                SOME ty =>
                                    let val () = check_exp_con ctx e ty
                                                 ("synth_exp_con: ExSumInj "
                                                 ^ "arg has wrong type")
                                    in
                                        c
                                    end
                              | NONE => raise Check ("synth_exp_con: "
                                                    ^ "no such sum label, "
                                                    ^ string_of_label lab))
                      | _ => raise Check
                             "synth_exp_con: ExSumInj into non-TySum type"
                end
          | ExCase (c_cod, e, labes) =>
                let val c_sum = synth_exp_con ctx e
                in
                    case whnf ctx c_sum of
                        TySum labtys =>
                            let val () =
                                ListPair.app
                                    (fn ((l1, c_dom), (l2, e)) =>
                                        let val () = check (label_eq (l1, l2))
                                                    (fn () => raise Check
                                                        ("synth_exp_con: label "
                                                        ^ "mismatch in ExCase"))
                                            val () = check_exp_con ctx e
                                                     (TyArrow (c_dom, c_cod))
                                                     ("synth_exp_con: type "
                                                     ^ "mismatch in ExCase")
                                        in
                                            ()
                                        end)
                                    (labtys, labes)
                            in
                                c_cod
                            end
                      | _ => raise Check
                             "synth_exp_con: ExCase over non-TySum type"
                end
          | ExTagInj (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf ctx c1 of
                        TyTag c2 =>
                            let val () = check_exp_con ctx e2 c2
                                         ("synth_exp_con: ExTagInj data has "
                                         ^ "incorrect type")
                            in
                                TyTagged
                            end
                      | _ => raise Check "synth_exp_con: ExTagInj into non-tag"
                end
          | ExNewTag c =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExNewTag carrying non-type"
                in
                    TyTag c
                end
          | ExIfTagOf (e1, e2, e3, e4) =>
                let val () = check_exp_con ctx e1 TyTagged
                             "synth_exp_con: ExIfTagOf on non-TyTagged"
                    val c_data =
                        case whnf ctx (synth_exp_con ctx e2) of
                            TyTag c => c
                          | _ => raise Check ("synth_exp_con: ExIfTagOf "
                                             ^ "comparison exp not a TyTag")
                    val c_res =
                        case whnf ctx (synth_exp_con ctx e3) of
                            TyArrow (c_data', c) =>
                                let val () = check_equiv_con ctx c_data c_data'
                                             KndType
                                             ("synth_exp_con: ExIfTagOf `then' "
                                             ^ "function has incorrect domain")
                                in
                                    c
                                end
                          | _ => raise Check ("synth_exp_con: ExIfTagOf `then' "
                                             ^ "exp not a function")
                    val () = check_exp_con ctx e4 c_res
                             ("synth_exp_con: ExIfTagOf `else' exp "
                             ^ "has incorrect type")
                in
                    c_res
                end
          | ExConAbs (cx, k, e) =>
                let val () = check_kind ctx k
                    val (cx, e) = alpha_vary (cx, e) replace_con_exp
                    val ctx = Ctx.extend ctx (DecConKind (cx, k))
                    val c = synth_exp_con ctx e
                in
                    TyForall (cx, k, c)
                end
          | ExConApp (e, c_arg) =>
                let val c_fun = synth_exp_con ctx e
                in
                    case whnf ctx c_fun of
                        TyForall (cx, k, c) =>
                            let val () = check_con_kind ctx c_arg k
                                         ("synth_con_kind: ExConApp arg "
                                         ^ "has wrong kind")
                            in
                                subst_con_con c_arg cx c
                            end
                      | _ => raise Check "con-application of a non-forall"
                end
          | ExLet (ex, e, e') =>
                let val c = synth_exp_con ctx e
                    val (ex, e') = alpha_vary (ex, e') replace_exp_exp
                    val ctx = Ctx.extend ctx (DecExpCon (ex, c))
                    val c' = synth_exp_con ctx e'
                in
                    c'
                end
          | ExModLet (mx, m, e, c) =>
                let val sg = synth_mod_sig ctx m
                    val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExModLet body type not a type"
                    val (mx, e) = alpha_vary (mx, e) replace_mod_exp
                    val ctx = Ctx.extend ctx (DecModSig (mx, sg))
                    val () = check_exp_con ctx e c
                             "synth_exp_con: ExModLet body type incorrect"
                in
                    c
                end
          | ExModProj (modv, lab, labs) =>
                let val sdecs =
                        case (synth_mod_sig ctx modv) of
                            SigStruct sds => sds
                          | _ => raise Check ("synth_exp_con: "
                                 ^ "module projection from a non-structure")
                    (* belt and suspenders check *)
                    val () = if module_value modv then
                                ()
                             else
                                raise Error "projection from a non-value module"
                    val sdecs = selfify modv sdecs labs
                in
                    case Ctx.sdecs_lookup_explab modv sdecs lab of
                        NONE => raise Check ("no such module label, "
                                            ^ string_of_label lab)
                      | SOME c => c
                end

    and check_exp_con ctx exp con s =
        let val con' = synth_exp_con ctx exp
        in
            equiv_con ctx con con' KndType
            handle Equiv => raise Check s
        end
  end
