signature IL3  =
  sig
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var

    datatype kind =
        KndType
      | KndArrow of kind * kind (* convar bound in kind2 *)
      | KndUnit
      | KndPair of kind * kind (* convar bound in kind2 *)

    datatype con =
        ConVar of convar
      | ConAbs of convar * kind * con
      | ConApp of con * con
      | ConUnit
      | ConPair of con * con
      | ConProj1 of con
      | ConProj2 of con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)

      | TyLit of Literal.lit_ty
      | TyArrow of con * con
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of con list
      | TyProd of con list
      | TyForall of convar * kind * con (* convar bound in con *)

    datatype exp =
        ExVar of expvar
      | ExLit of Literal.lit
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
            (* includes projection out of bundle*)
      | ExFix of int * (expvar * expvar * con * con * exp) list
      | ExApp of exp * exp
      | ExHandle of exp * (expvar * exp) (* expvar bound in exp2 *)
      | ExRaise of con * exp
      | ExRef of exp
      | ExGet of exp
      | ExSet of exp * exp
      | ExRoll of con * exp
      | ExUnroll of exp
      | ExProd of exp list
      | ExProj of int * exp
      | ExSumInj of con * int * exp
      | ExCase of con * exp * exp list
      | ExTagInj of exp * exp
      | ExNewTag of con
      | ExIfTagOf of exp * exp * exp * exp
      | ExConAbs of convar * kind * exp (* convar bound in exp *)
      | ExConApp of exp * con
      | ExLet of expvar * exp * exp (* expvar bound in exp2 *)
    type fbnd = (expvar * expvar * con * con * exp)


    (*** derived forms ***)

    val kndProd : kind list -> kind
    val conProd : con list -> con
    val conProj : int * con -> con

    val exLam : expvar * con * con * exp -> exp

    val tyUnit : con
    val tyBool : con
    val tyInt : con
    val tyChar : con
    val tyString : con

    val exUnit : exp
    val exBool : bool -> exp
    val exInt : int -> exp
    val exChar : char -> exp
    val exString : string -> exp

    val exIf : con -> exp * exp * exp -> exp


    (*** mapping ***)

    val map_con : (convar -> con)
               -> (convar * con -> convar * con)
               -> con -> con

    val map_exp : (expvar -> exp)
               -> (fbnd list -> fbnd list)
               -> (expvar * exp -> expvar * exp)
               -> (convar * exp -> convar * exp)
               -> (con -> con)
               -> exp -> exp


    (*** replacement of convars ***)

    val replace_con_con : convar -> convar -> con -> con
    val replace_con_exp : convar -> convar -> exp -> exp

    (*** replacement of expvars ***)

    val replace_exp_exp : expvar -> expvar -> exp -> exp


    (*** alpha varying ***)

    val alpha_vary : 'a Variable.var * 'b
                  -> ('a Variable.var -> 'a Variable.var -> 'b -> 'b)
                  -> 'a Variable.var * 'b

    val alpha_vary2 : 'a Variable.var * 'b -> 'a Variable.var * 'b
                   -> ('a Variable.var -> 'a Variable.var -> 'b -> 'b)
                   -> 'a Variable.var * ('b * 'b)

    val alpha_vary_many : 'a Variable.var list * 'b list
                       -> ('a Variable.var -> 'a Variable.var -> 'b -> 'b)
                       -> 'a Variable.var list * 'b list


    (*** substitution for convars ***)

    val subst_con_con : con -> convar -> con -> con
    val subst_con_exp : con -> convar -> exp -> exp
  end
