signature IL4 =
  sig
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var

    datatype kind =
        KndType
      | KndArrow of kind * kind (* convar bound in kind2 *)
      | KndUnit
      | KndPair of kind * kind (* convar bound in kind2 *)

    datatype con =
        ConVar of convar
      | ConAbs of convar * kind * con
      | ConApp of con * con
      | ConUnit
      | ConPair of con * con
      | ConProj1 of con
      | ConProj2 of con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)

      | TyLit of Literal.lit_ty
      | TyCont of con (* instead of TyArrow *)
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of con list
      | TyProd of con list
      | TyForall of convar * kind * con (* convar bound in con *)

    datatype value =
        ValVar of expvar
      | ValLit of Literal.lit
      (* includes bundle projection *)
      | ValFix of int * (expvar * expvar * con * exp) list 
      | ValProd of value list
      | ValSumInj of con * int * value
      | ValTagInj of value * value
      | ValConAbs of convar * kind * value (* convar bound in value *)
      | ValConApp of value * con
      | ValRoll of con * value

    and exp =
        ExHalt of value
      | ExApp of value * value
      | ExCase of value * (expvar * exp) list
      | ExIfTagOf of value * value * (expvar * exp) * exp
                                   (* expvar bound in exp1 *)
      | ExLet of bnd * exp (* bnd's expvar bound in exp2 *)

    and bnd =
        BndVal of expvar * value
      | BndUnOp of expvar * Literal.unop * value
      | BndBinOp of expvar * value * Literal.binop * value
      | BndRef of expvar * value
      | BndGet of expvar * value
      | BndSet of expvar * value * value
      | BndProj of expvar * int * value
      | BndNewTag of expvar * con
      | BndUnroll of expvar * value

    type fbnd = (expvar * expvar * con * exp)
              (* fix f       (x      : t) => e   end *)


    (*** derived forms ***)

    val kndProd : kind list -> kind
    val conProd : con list -> con
    val conProj : int * con -> con

    val tyUnit : con
    val tyBool : con

    val tyInt : con
    val tyChar : con
    val tyString : con

    val valUnit : value
    val valInt : int -> value
    val valChar : char -> value
    val valString : string -> value

    val valLam : expvar * con * exp -> value


    (*** mapping ***)

    val map_con : (convar -> con)
               -> (convar * con -> convar * con)
               -> con -> con

    val map_value : (expvar -> value)
                 -> (fbnd list -> fbnd list)
                 -> (convar * value -> convar * value)
                 -> (con -> con)
                 -> value -> value

    val map_exp : (expvar * exp -> expvar * exp)
               -> (bnd * exp -> bnd * exp)
               -> (value -> value)
               -> exp -> exp

    val map_bnd : (value -> value)
               -> (con -> con)
               -> bnd -> bnd


    (*** replacement of convars ***)

    val replace_con_con : convar -> convar -> con -> con
    val replace_con_value : convar -> convar -> value -> value
    val replace_con_exp : convar -> convar -> exp -> exp
    val replace_con_bnd : convar -> convar -> bnd -> bnd

    (*** replacement of expvars ***)

    val replace_exp_exp : expvar -> expvar -> exp -> exp


    (*** alpha varying ***)

    val alpha_vary : 'a Variable.var * 'b
                  -> ('a Variable.var -> 'a Variable.var -> 'b -> 'b)
                  -> 'a Variable.var * 'b

    val alpha_vary2 : 'a Variable.var * 'b -> 'a Variable.var * 'b
                   -> ('a Variable.var -> 'a Variable.var -> 'b -> 'b)
                   -> 'a Variable.var * ('b * 'b)

    val alpha_vary_bnd_exp : bnd * exp -> bnd * exp


    (*** substitution of convars ***)

    val subst_con_con : con -> convar -> con -> con
  end
