signature IL6 =
  sig
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var
    
    datatype kind =
        KndType
      | KndArrow of kind * kind (* convar bound in kind2 *)
      | KndUnit
      | KndPair of kind * kind (* convar bound in kind2 *)
    
    datatype con =
        ConVar of convar
      | ConAbs of convar * kind * con
      | ConApp of con * con
      | ConUnit
      | ConPair of con * con
      | ConProj1 of con
      | ConProj2 of con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)

      | TyLit of Literal.lit_ty
      | TyCont of (convar * kind) list * con (* instead of (TyCont c)*)
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of con list
      | TyProd of con list
      | TyForall of convar * kind * con (* convar bound in con *)
      | TyExists of convar * kind * con
 
    datatype word = 
        WProd of value list
      | WRoll of con * value
      | WSumInj of con * int * value
      | WPack of ((con * value) * con)
      | WTagInj of value * value
      | WRef of value
      | WNewTag of con

    and value =
        ValVar of expvar
      | ValLit of Literal.lit
      | ValConAbs of convar * kind * value
      | ValConApp of value * con
      | ValLet of bnd * value
      
    and exp =
        ExHalt of value
      | ExApp of value * value
      | ExCase of value * (expvar * exp) list
      | ExIfTagOf of value * value * (expvar * exp) * exp
                                   (* expvar bound in exp1 *)
      | ExLet of bnd * exp (* bnd's expvar bound in exp2 *)
      | ExUnpack of convar * expvar * value * exp (* vars bound in exp *)
    
    and bnd =
        BndVal of expvar * value
      | BndAlloc of expvar * word
      | BndUnOp of expvar * Literal.unop * value
      | BndBinOp of expvar * value * Literal.binop * value
      | BndGet of expvar * value
      | BndSet of expvar * value * value
      | BndProj of expvar * int * value
      | BndUnroll of expvar * value
    type fbnd = ((convar * kind) list * expvar * con * exp)
                    (* Lam [a : k].lam x : c. e *)

    type program = (expvar * fbnd) list * exp
    
    (*** derived forms ***)

    val kndProd : kind list -> kind
    val conProd : con list -> con
    val conProj : int * con -> con

    val tyUnit : con
    val tyBool : con

    val tyInt : con
    val tyChar : con
    val tyString : con

    val valUnit : value
    val valInt : int -> value
    val valChar : char -> value
    val valString : string -> value
    
    (*** mapping ***)

    val map_con : (convar -> con)
               -> (convar * con -> convar * con)
               -> con -> con

    val map_word : (value -> value)
                -> (con -> con)
                -> word -> word

    val map_value : (expvar -> value)
                 -> (convar * value -> convar * value)
                 -> (bnd * value -> bnd * value)
                 -> (con -> con)
                 -> value -> value

    val map_exp : (expvar * exp -> expvar * exp)
               -> (convar * expvar * exp -> convar * expvar * exp)
               -> (bnd * exp -> bnd * exp)
               -> (value -> value)
               -> exp -> exp

    val map_bnd : (value -> value)
               -> (word -> word)
               -> (con -> con)
               -> bnd -> bnd

    (*** replacement of convars ***)

    val replace_con_con : convar -> convar -> con -> con
    val replace_con_value : convar -> convar -> value -> value
    val replace_con_exp : convar -> convar -> exp -> exp
    val replace_con_bnd : convar -> convar -> bnd -> bnd
    
    (*** replacement of expvars ***)

    val replace_exp_exp : expvar -> expvar -> exp -> exp
    val replace_exp_value : expvar -> expvar -> value -> value
    
    (*** alpha varying ***)

    val alpha_vary : 'a Variable.var * 'b
                  -> ('a Variable.var -> 'a Variable.var -> 'b -> 'b)
                  -> 'a Variable.var * 'b

    val alpha_vary2 : 'a Variable.var * 'b -> 'a Variable.var * 'b
                   -> ('a Variable.var -> 'a Variable.var -> 'b -> 'b)
                   -> 'a Variable.var * ('b * 'b)

    val alpha_vary_bnd_exp : bnd * exp -> bnd * exp

    (*** substitution of convars ***)

    val subst_con_con : con -> convar -> con -> con

  end
