signature IL7 =
  sig
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var
    
    datatype value =
        ValVar of expvar
      | ValLit of Literal.lit

    and bnd =
        BndLit of expvar * Literal.lit
      | BndUnOp of expvar * Literal.unop * value
      | BndBinOp of expvar * value * Literal.binop * value
      | BndAlloc of expvar * int
      | BndNewTag of expvar
      | BndProj of expvar * int * value
      | BndSet of expvar * int * value (* x[i] <- v *)

    datatype exp =
        ExHalt of value
      | ExApp of value * value
          (* switch v of e1, ..., en *)
      | ExSwitch of value * exp list
      | ExLet of bnd * exp

    type fbnd =  expvar * exp (* lam x.e *)
    (* letrec f1 = fb1, ..., fn = fbn in e *)
    type program = (expvar * fbnd) list * exp 

    val map_value : (expvar -> value) -> value -> value
    val map_bnd : (expvar -> expvar)
               -> (value -> value)
               -> bnd -> bnd
    val map_exp : (value -> value)
               -> (bnd * exp -> bnd * exp)
               -> exp -> exp

    val replace_exp_exp : expvar -> expvar -> exp -> exp
    val replace_exp_value : expvar -> expvar -> value -> value
    val replace_exp_bnd : expvar -> expvar -> bnd -> bnd
    
    (* derived forms *)
    val valInt : int -> value
    val valChar : char -> value
    val valString : string -> value
    val bndUnit : expvar -> bnd
  end
