structure IL7 :> IL7 =
  struct
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var
    
    datatype value =
        ValVar of expvar
      | ValLit of Literal.lit

    and bnd =
        BndLit of expvar * Literal.lit
      | BndUnOp of expvar * Literal.unop * value
      | BndBinOp of expvar * value * Literal.binop * value
      | BndAlloc of expvar * int
      | BndNewTag of expvar
      | BndProj of expvar * int * value
      | BndSet of expvar * int * value (* x[i] <- v *)

    datatype exp =
        ExHalt of value
      | ExApp of value * value
          (* switch v of e1, ..., en default ed *)
      | ExSwitch of value * exp list
      | ExLet of bnd * exp

    type fbnd = expvar * exp (* lam x.e *)
    (* letrec f1 = fb1, ..., fn = fbn in e *)
    type program = (expvar * fbnd) list * exp

    fun map_value fvar v =
      case v of
        ValVar x => fvar x
      | x => x


    fun map_bnd fvar fval bnd =
      case bnd of
        BndUnOp(x, oper, v) => BndUnOp(x, oper, fval v)
      | BndBinOp(x, v1, oper, v2) => BndBinOp(x, fval v1, oper, fval v2)
      | BndProj(x, i, v) => BndProj(x, i, fval v)
      | BndSet(x, i, v) => BndSet(fvar x, i, fval v)
      | x => x

    fun map_exp fval fbndexp exp =
      case exp of
        ExHalt v => ExHalt(fval v)
      | ExApp(v1, v2) => ExApp(fval v1, fval v2)
      | ExSwitch(v, es) => ExSwitch(fval v, map (map_exp fval fbndexp) es)
      | ExLet be => ExLet (fbndexp be)
    
    fun bnd_var bnd =
        case bnd of
            BndLit (x, _) => x
          | BndUnOp (x, _, _) => x
          | BndBinOp (x, _, _, _) => x
          | BndAlloc (x, _) => x
          | BndSet (x, _, _) => x
          | BndProj (x, _, _) => x
          | BndNewTag x => x

    fun replace_exp_value expvar expvar' =
      map_value (fn x => if Variable.eq(expvar, x)
                         then ValVar expvar'
                         else ValVar x)
    fun replace_exp_bnd expvar expvar' =
      map_bnd (fn x => if Variable.eq(expvar, x)
                       then expvar'
                       else x)
              (replace_exp_value expvar expvar')

    fun replace_exp_exp expvar expvar' =
      map_exp (replace_exp_value expvar expvar')
              (fn (b, e) =>
                let val b = replace_exp_bnd expvar expvar' b
                    val x = bnd_var b
                in
                    if Variable.eq(x, expvar)
                    then (b, e)
                    else (b, replace_exp_exp expvar expvar' e)
                end)

    (* derived forms *)
    fun valInt i = ValLit (Literal.LitInt i)
    fun valChar c = ValLit (Literal.LitChar c)
    fun valString s = ValLit (Literal.LitString s)
    fun bndUnit x = BndAlloc(x, 0)
  end
