structure IL7Print =
struct
  open IL7
  
  val $ = Layout.str
  val % = Layout.mayAlign
  val & = Layout.seq

  fun var v = $(Variable.toString v)
  
  (* values *)
  fun literal_to_layout lit =
    case lit of
        Literal.LitInt i => $(Int.toString i)
      | Literal.LitChar c => $("'" ^ Char.toString c ^ "'")
      | Literal.LitString s => $("\"" ^ s ^ "\"")
  
  fun value_to_layout value =
    case value of
        ValVar x => var x
      | ValLit lit => literal_to_layout lit
  
  and bnd_to_layout bnd =
      let fun layb (x, l) = %[ &[var x, $" ="], Layout.indent 4 l ]
      in
          case bnd of
              BndLit (x, l) => layb (x, literal_to_layout l)
            | BndAlloc(x, i) =>
                  layb (x, %[ $"alloc", $(Int.toString i)])
            | BndProj (x, i, v) =>
                  layb (x, %[ &[$"#", $(Int.toString i)],
                              value_to_layout v])
            | BndUnOp (x, oper, v) =>
                  layb (x, %[$(Literal.unop_to_string oper),
                              value_to_layout v])
            | BndBinOp (x, v1, oper, v2) =>
                  layb (x, %[value_to_layout v1,
                             &[$(Literal.binop_to_string oper), $" ",
                               value_to_layout v2]])
            | BndSet (x, i, v) =>
                   &[ var x, $"[", $(Int.toString i), $"] <- ",
                              value_to_layout v]
            | BndNewTag x => layb (x, $"newtag")
      end
  
  and atomic_exp_to_layout exp =
      case exp of
          _ => Layout.paren (exp_to_layout exp)

  and app_exp_to_layout exp =
      case exp of
          ExHalt v => %[$"halt", value_to_layout v]
        | ExApp (v1, v2) => %[value_to_layout v1,
                              value_to_layout v2]
        | _ => atomic_exp_to_layout exp
  and exp_to_layout exp =
        case exp of
            ExSwitch (v, xexps) =>
                Layout.align
                [ &[$"switch ", value_to_layout v, $" of"],
                  Layout.indent 4 (Layout.sequence ("{", "}", ",")
                                    (map exp_to_layout xexps))]
          | ExLet (b, e) =>
                %[ &[$"let ", bnd_to_layout b],
                   $"in",
                   (exp_to_layout e) ]
          | _ => app_exp_to_layout exp
      
      fun named_fbnd_to_layout (f, (x, e)) =
            %[ &[var f, $" = ", $"lam ", var x, $"."],
               exp_to_layout e]

      fun program_to_layout (fs, e) =
        %[Layout.sequence ("letrec ", "", ",") (map named_fbnd_to_layout fs),
          &[$"in ", exp_to_layout e]]
end
