#include "runtime.h"
#include "gc.h"

#define DEFAULT_HEAP_SIZE 10000

/* Makes sure static pointers don't get traversed by the gc. */
#define safe_ptr(x) ((hotc_t)(x.num | 1))

/* For memory management */
hotc_t* allocp; /* current start of mem */
hotc_t* limitp; /* current end of mem */

hotc_t exception_tag_counter = (hotc_t)0; /* Counter for exception tags */

/* Constants for true, false, and unit to be used for equals*/
static hotc_t hotc_unit_body[] = {0, (hotcint_t)NULL};
static hotc_t hotc_unit = (hotc_t)((union hotc_t*)&hotc_unit_body);
static hotc_t hotc_true_body[] = {2, 3, (hotcint_t)&hotc_unit_body};
static hotc_t hotc_false_body[] = {2, 5, (hotcint_t)&hotc_unit_body};
static hotc_t hotc_true = (hotc_t)((union hotc_t*)&hotc_true_body);
static hotc_t hotc_false = (hotc_t)((union hotc_t*)&hotc_false_body);

/* Compiled main function */
hotcfun_t* hotc_mainfun(hotc_t* arg);

void hotc_check(hotc_t* root, hotcint_t size) {
  if(allocp + size < limitp)
    return;
  else {
    gc_collect(root, &allocp, &limitp);
    if(allocp + size >= limitp)
    {
      fprintf(stderr, "out of memory...aborting.\n");
      exit(1);
    }
  }
}

hotc_t* hotc_alloc(hotcint_t size, int donttrace) {
    hotc_t* result = allocp;
    allocp = allocp + size + 1;
    result[0].num = size << 1 | ((!!donttrace) << 31);
    return result;
}

void hotc_halt(hotc_t result) {
  exit(0);
}

hotc_t concat(hotc_t x, hotc_t y) {
    hotcstr_t str1 = unbox_str(x);
    hotcstr_t str2 = unbox_str(y);
    hotcint_t len1 = strlen(str1);
    hotcint_t len2 = strlen(str2);
    if(len1 + len2 > MAX_STR_SIZE) {
      fprintf(stderr, "exceeded string maximum...exiting\n");
      exit(1);
    }
    hotc_t out = box_ptr(hotc_alloc(len1 + len2 + 1, TRUE));
    strcpy(unbox_str(out), str1);
    strcat(unbox_str(out), str2);
    return (hotc_t)out;
}

hotc_t int_to_str(hotc_t x) {
  hotc_t out = box_ptr(hotc_alloc((MAX_STR_SIZE + 3) / 4, TRUE));
  sprintf(unbox_str(out), "%d", unbox_int(x));
  return out;
}

hotc_t chr_to_str(hotc_t x) {
  hotc_t out = box_ptr(hotc_alloc(1 , TRUE));
  sprintf(unbox_str(out), "%c", unbox_chr(x));
  return out;
}

hotc_t equals(hotc_t x, hotc_t y) {
  return x.num == y.num ? safe_ptr(hotc_true) : safe_ptr(hotc_false);
}

int main(int argc, char** argv) {
  void* curfun = hotc_mainfun;
  hotc_t arg = safe_ptr(hotc_unit);
  int size = DEFAULT_HEAP_SIZE;
  if(argc > 2)
  {
    printf("usage: %s <heapsize>\n");
    exit(1);
  }
  else if(argc == 2)
    size = atoi(argv[1]);
  if(gc_init(size, &allocp, &limitp))
  {
    fprintf(stderr, "gc failed to initialize...aborting.\n");
    exit(1);
  }
  while(1)
  {
    curfun = ((hotcfun_t*)curfun)(&arg);
  }
  return -1; /* Should never return */
}
