#ifndef __RUNTIME__
#define __RUNTIME__

#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#define FALSE 0
#define TRUE 1

/* Result codes */
#define SUCCESS 0
#define SEGFAULT 1
#define OUTOFMEM 2
#define OUTOFTIME 3
#define INITIALIZATION 4
#define CONSISTENCY 5

/* Maximum size for strings. */
#define MAX_STR_SIZE 1024

/* We assume we are on a 32-bit architecture */
#define HIGH_BIT_MASK (1 << 31)

typedef int32_t hotcint_t;
/* Needs to be bigger than just a char so we can box it.
 * Not actually wasting space either since we coerece to hotc_t anyway */
typedef uint32_t hotcchr_t;
typedef char* hotcstr_t;

/* We really hope that this is word sized */
union hotc_t {
    hotcint_t num;
    hotcstr_t str;
    hotcchr_t chr;
    union hotc_t* arr;
    void* fun;
};
typedef union hotc_t hotc_t;

typedef void* hotcfun_t(hotc_t* arg);

hotc_t* hotc_alloc(int size, int donttrace);
void hotc_check(hotc_t* root, int size);
void hotc_halt(hotc_t result);

#define arr_size(p) ((((p)[0]).num & ~HIGH_BIT_MASK) >> 1)

#define box_int(x) ((hotc_t)(((x) << 1) | 1))
#define box_chr(x) ((hotc_t)(((x) << 1) | 1))
#define box_fun(x) ((hotc_t)(((hotcint_t)x) | 1))
#define box_ptr(x) ((hotc_t)(((hotcint_t)(x))))

#define unbox_int(x) (((x).num >> 1))
#define unbox_chr(x) (((x).chr >> 1))
#define unbox_fun(x) ((hotcfun_t*)(((x).num) & ~1))
#define unbox_ptr(x) ((hotc_t*)((x).num & ~1))
#define unbox_str(x) ((char*)((x).arr + 1))

#define add(x, y) ((hotc_t)((x).num + (y).num - 1))
#define sub(x, y) ((hotc_t)((x).num - (y).num + 1))
#define mul(x, y) (box_int(unbox_int(x) * unbox_int(y)))
hotc_t equals(hotc_t x, hotc_t y);
hotc_t concat(hotc_t x, hotc_t y);
#define print(x) ((hotc_t)(box_int(printf(unbox_str(x)))))
hotc_t int_to_str(hotc_t x);
hotc_t chr_to_str(hotc_t x);


#define str_to_int(x) (atoi(unbox_str(x)))
#define chr_to_int(x) (x)
#define int_to_chr(x) (x)
#define app(x, y) { \
                      *arg = y; \
                      return unbox_fun(x); \
                  }
#define get(x, i) ((unbox_ptr(x))[i])
#define set(x, i, v) ((unbox_ptr(x))[i] = (hotc_t)(v))

#endif
