/* Constructs a simple two node loop in memory */
/* Run with around 450 bytes of memory to make sure
 * the GC gets invoked.
 */
#include <runtime.h>

extern hotcint_t exception_tag_counter;
void* hotc_mainfun(hotc_t* arg);

void* hotc_mainfun(hotc_t* arg)
{
  int x;
  hotc_check(arg, 6); /* head + tail + 2 header words */
  hotc_t head = (hotc_t)hotc_alloc(2, FALSE); /* value and next pointer */
  hotc_t tail = (hotc_t)hotc_alloc(2, FALSE); /* value and next pointer */
  set(head, 1, box_int(42));
  set(head, 2, tail);
  set(tail, 1, box_int(24));
  set(tail, 2, head);
  /* force a gc by allocating some memory we don't use*/
  hotc_check(&head, 100);
  hotc_t unused = (hotc_t)hotc_alloc(100, FALSE);
  hotc_check(&head, 100);

  /* check consistency */
  hotcint_t val1 = unbox_int(get(head, 1));
  hotcint_t val2 = unbox_int(get(tail, 1));
  if(val1 != 42 && val2 != 24) {
    printf("memory no longer correct\n");
    exit(CONSISTENCY);
  }
  tail = get(head, 2);

  /* check that the structure is still valid */
  hotc_t head2 = get(tail, 2);
  hotc_t tail2 = get(head, 2);
  val1 = unbox_int(get(head2, 1));
  val2 = unbox_int(get(tail2, 1));
  if(val1 != 42 && val2 != 24) {
    printf("memory no longer loops correctly\n");
    exit(CONSISTENCY);
  }
  /* check that a change propagates */
  set(head, 1, box_int(84));
  set(tail, 1, box_int(12));
  val1 = unbox_int(get(head2, 1));
  val2 = unbox_int(get(tail2, 1));
  if(val1 != 84 && val2 != 12) {
    printf("memory changes don't propogate\n");
    exit(CONSISTENCY);
  }
  /* check that a change propagates */

  printf("success\n");
  hotc_halt(*arg);
}
